#ifndef RTCP_TEST
#include "defines.h"
#endif

#include <stdint.h>

#if RTCP_FEEDBACK
#ifndef _RTCP_H_
#define _RTCP_H_

/***************************/

#define RTCP_MIN_TIME 0                        /* P2P, no multiparty */
#define RTCP_BW_FRACTION 0.05
#define RTCP_SENDER_BW_FRACTION 0.25           
#define RTCP_RCVR_BW_FRACTION  (1-RTCP_SENDER_BW_FRACTION);
#define RTCP_PROBABLE_RTCP_SIZE 64           /* bytes, RR with one block +
                                              * SDES with cname stian@master.ntnu.no */ 
#define RTCP_T_RR_INTERVAL 5                 /* seconds */
#define RTCP_ENABLE_TIMER_RECONSIDERATION 0

#define RTCP_T_DITHER_MAX 0.15           /* max dither for FB in seconds */
#define RTCP_T_MAX_FB_DELAY 60.0         /* max delay for FB to be useful in seconds
                                          * we assume that they're always
                                          * useful */

#define RTCP_SSRC_DECODER 0x87654321    /* constant ssrc chosen for simplicity */
#define RTCP_SSRC_ENCODER 0x12345678      /* constant ssrc also in encoder */
#define RTCP_MAX_PACKET_SIZE (65536 - 28) /* Max size IP packet - IP header - UDP header */
#define RTCP_MAX_SDES_ITEM_LENGTH 25

/***************************/

#ifndef RTCP_TEST
#include "rtp.h"
#else
typedef unsigned char byte;
typedef struct 
{
  unsigned int v;          //!< Version, 2 bits, MUST be 0x2
  unsigned int p;          //!< Padding bit, Padding MUST NOT be used
  unsigned int x;          //!< Extension, MUST be zero
  unsigned int cc;         /*!< CSRC count, normally 0 in the absence 
                                of RTP mixers */
  unsigned int m;          //!< Marker bit
  unsigned int pt;         //!< 7 bits, Payload Type, dynamically established 
  uint16_t  seq;        /*!< RTP sequence number, incremented by one for
                                each sent packet */
  uint16_t old_seq;    //!< to detect wether packets were lost
  unsigned int timestamp;  //!< timestamp, 27 MHz for H.264
  unsigned int ssrc;       //!< Synchronization Source, chosen randomly
  byte *       payload;    //!< the payload including payload headers
  unsigned int paylen;     //!< length of payload in bytes
  byte *       packet;     //!< complete packet including header and payload
  unsigned int packlen;    //!< length of packet, typically paylen+12
} RTPpacket_t;
#endif

//#define DEFAULT_ENCODER_HOST "127.0.0.1"
//#define DEFAULT_ENOCDER_RTCP_PORT 62003


typedef enum {
  
  
  /* RTCP packet types according to
   * http://www.iana.org/assignments/rtp-parameters */
  RTCP_FIR   = 192,
  RTCP_NACK  = 193,
  RTCP_SR    = 200,
  RTCP_RR    = 201,
  RTCP_SDES  = 202,
  RTCP_BYE   = 203,
  RTCP_APP   = 204,
  RTCP_RTPFB = 205,
  RTCP_PSFB  = 206,
  RTCP_XR    = 207
} rtcp_packet_types;


typedef struct {
  uint32_t ssrc_n;                 /* ssrc, 32 bits */
  uint8_t fl;                    /* fraction lost, 8 bits */
  uint32_t cnlp;                   /* cumulative number of lost packets,
                                         * 24 bits */
  uint32_t ehsn;                   /* extended highest seq no, 32 bits */
  uint32_t jitter;                 /* interarrival jitter, 32 bits */
  uint32_t lsr;                    /* last SR, 32 bits */
  uint32_t dlsr;                        /* dealy since last SR, 32 bits */
} rtcp_rr_report_block_t;


typedef struct {
  uint8_t v;                     /* version, 2 bits */
  uint8_t p;                     /* padding, 1 bit */
  uint8_t rc;                    /* rr count, 5 bits */
  uint8_t pt;                    /* packet type, 2 bits */
  uint16_t length;                  /* length of packet, 16 bits */
  uint32_t ssrc;                   /* ssrc, 32 bits */
  rtcp_rr_report_block_t *rb;           /* report block 1. According to the
                                         * standard this should be an array,
                                         * but we only need one. */
} rtcp_rr_t;

/* An extremely simplified struct for a SDES message which only can contain
 * one SDES item, namely the mandatory CNAME item. */
typedef struct {
  uint8_t v;
  uint8_t p;
  uint8_t sc;
  uint8_t pt;
  uint16_t length;
  uint32_t chunk_csrc;
  uint8_t chunk_item_type;
  uint8_t chunk_item_length;
  char *chunk_item_cname;                
} rtcp_sdes_t;

typedef struct {
  uint16_t pid;                     /* RTP seq of lost packet, 16 bits */
  uint16_t blp;                     /* Bitmask of following lost packets,
                                         * 16 bits */
} rtcp_fci_nack_t;

typedef struct {
  uint8_t v;                            /* version, 2 bits */
  uint8_t p;                            /* padding, 1 bit */
  uint8_t fmt;                          /* feedback message type, 5 bits */
  uint8_t pt;                           /* payload type, 8 bits */
  uint16_t length;                      /* length of packet, 16 bits */
  uint32_t ssrc_sender;                 /* ssrc of the sender of this packet,
                                         * 32 bits*/
  uint32_t ssrc_source;                 /* ssrc of the related media source,
                                         * 32 bits */
  rtcp_fci_nack_t *fci;               /* list feedback control information */
  int n_fci;                            /* coint elements in fci list */
} rtcp_fb_t;

typedef struct {
  rtcp_rr_t *rr;
  rtcp_sdes_t *sdes;
  rtcp_fb_t *fb;                        /* Since we only have one type of FB,
                                           only one fb entry is required.*/
  byte *packet;
  int packlen;
} rtcp_packet_t;


typedef struct {
  /* These are the values we need from struct source in RFC 3550 */
  uint16_t max_seq;
  uint16_t base_seq;
  uint32_t cycles;
  uint32_t received;
  uint32_t expected_prior;
  uint32_t received_prior;
  
//  unsigned long  SR_received;   /* SR records received */
//  unsigned long  dest_SSRC;     /* SSRC of first source */
//  unsigned long  pkt_count;     /* count of packets received */
//  unsigned long  octet_count;   /* total payload octets from sender */
//  unsigned long  pkt_lost;      /* cumulative count of packets lost */
//  double fraction_lost; /* count of fractional packets lost */
//  unsigned long  highest_seq_no;/* highest sequence number found */
//  unsigned long  jitter;        /* inter arrival jitter calculated */
//  unsigned long  last_SR;       /* last SR record received */
//  unsigned long  delay_since_last_SR; /* delay since last SR record received */
//  unsigned long  delay_since_last_RR; /* delay since last RR record received */

//  unsigned long avg_pkt_size;
//  unsigned long sent_pkt_size;
} rtcp_stats_t;

typedef struct {
  char encoder_host[256];
  unsigned short encoder_port;
  struct sockaddr_in encoder_addr;
  int sockfd_out;
} rtcp_connection_t;




short rtcp_alive;

void rtcp_rtp_packet_lossed (uint16_t seq);
void rtcp_rtp_packet_received (RTPpacket_t *p);
void rtcp_init (uint16_t firstseq);

/* Initialize RTCP
 *
 * Parameters: Feedback mode, encoder addr, encoder port
 * 
 * Initialies div stuff
 */

/* Report packet
 *
 * Parameters: seq no, timestamp, lost?
 *
 * Update statistics and variables. Decide if SR / FB should be sendt.
 *
 */

#endif
#endif
