#ifndef RTCP_TEST
#include "defines.h"
#include "global.h"
#endif

#if RTCP_FEEDBACK

#include <stdlib.h>
#include <sys/timeb.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <unistd.h>
#include <float.h>

#include "rtcp.h"
#include "schedule.h"


typedef struct {
  int id;
} rtcp_send_rr_maybe_arg;

typedef struct {
  int id;
} rtcp_send_early_fb_arg;

static int rtcp_send_packet(rtcp_packet_t *, rtcp_connection_t *);
//static void rtcp_send_rr_maybe (void *);
static void rtcp_send_rr_maybe(void);
static void rtcp_send_early_fb(void);
static void rtcp_send_early_fb_maybe(uint16_t);
//static void rtcp_schedule_early_fb(double);
static double rtcp_interval();
static void rtcp_add_fb(uint16_t);
static void rtcp_new_fb(uint16_t, double);
static void rtcp_set_rr(rtcp_packet_t *);
static void rtcp_set_sdes(rtcp_packet_t *);
static void rtcp_set_fb(rtcp_packet_t *);
static int rtcp_compose_packet(rtcp_packet_t *);

/* Global variables for RTCP. The naming convention introduced in RFC3550 and
 * RFC4585 is used to be able to understand the code more easily. */
static short senders;
static short members;
static short pmembers;
static double tp;
static double tn;
static double te;
static double t_rr_last;
//static double Tmin;
static short allow_early;
static short we_sent;
//static double rtcp_bw;
static short initial;
static double avg_rtcp_size;
static int next_rr_id;
static int next_fb_id;
static short rtcp_early_fb_waiting;
static short rtcp_rr_fb_waiting;
static short t_rr_interval;                /* seconds */

static rtcp_packet_t *rtcp_packet; 
static rtcp_connection_t rtcp_con;
static rtcp_stats_t rtcp_stats;
/* Implements (parts of) RTPC as defined in RFC3550 with the additions in
 * RFC4585. */

static pthread_mutex_t rtcp_mutex = PTHREAD_MUTEX_INITIALIZER;

#define TIME_NO_EARLY_FB DBL_MAX
#define TIME_NaN -1
#define RTCP_DEBUG_ENABLE 1

#if RTCP_DEBUG_ENABLE
#define RTCP_DEBUG(s) printf("%.3f: ", GET_TIME_SHORT(timenow())); printf s
#else
#define RTCP_DEBUG(s)
#endif

typedef enum {
  RTCP_LOSS_DETECTED = 1,
  RTCP_SCHEDULE_FB = 2,
  RTCP_ADD_FB_TO_FB = 3,
  RTCP_ADD_FB_TO_RR = 4,
  RTCP_LOSS_IGNORED = 5,
  RTCP_SEND_FB = 6,
  RTCP_SEND_RR = 7
} rtcp_event;

#define RTCP_PRINT_EVENT(s, e) \
  do { \
    printf("%.3f: <[%d]> ", GET_TIME_SHORT(timenow()), e); \
    printf s; \
    printf("\n"); \
  } while(0)

#define GET_TIME_SHORT(f) (((int) f % 100000) + (f - (int) f))

/** Helper functions */
#ifdef RTCP_TEST
static inline int max(int a, int b)
{
  return a > b ? a : b;
}
#endif

static inline double timenow(void)
{
  struct timeb t;
  ftime(&t); 
  return t.time + (double) t.millitm / 1000;
}

/*
 * Call this for every RTP packet received so the statistics are updated, that
 * includes also late and duplicate packets.
 */
void rtcp_rtp_packet_received (RTPpacket_t *p)
{
  //RTCP(("RTP packet %d successfully received\n", p->seq));
  rtcp_stats.received++;
  // todo: (low pri) handle cycles
  rtcp_stats.max_seq = max(p->seq, rtcp_stats.max_seq);
}


/*
 * Call this for every RTP packet that is lost
 * */
void rtcp_rtp_packet_lossed (uint16_t seq)
{
//  RTCP_DEBUG(("Loss reported on RTP packet %d\n", seq));
  RTCP_PRINT_EVENT(("Lost packet %hu", seq), RTCP_LOSS_DETECTED);
  rtcp_send_early_fb_maybe(seq);
}


/*
 * Allocate an RTP packet in memory
 * */
static void rtcp_packet_alloc(rtcp_packet_t **pp)
{
  rtcp_packet_t *p;
  
  if ((p = (rtcp_packet_t *) malloc(sizeof(rtcp_packet_t))) == NULL) {
    perror("rtcp_packet_alloc() malloc() #1");
    exit(-2);
  }

  if ((p->rr = (rtcp_rr_t *) malloc(sizeof(rtcp_rr_t))) == NULL) {
    perror("rtcp_packet_alloc() malloc() #2");
    exit(-2);
  }

  if ((p->rr->rb = (rtcp_rr_report_block_t *) malloc(sizeof(rtcp_rr_report_block_t))) == NULL) {
    perror("rtcp_packet_alloc() malloc() #3");
    exit(-2);
  }

  if ((p->sdes = (rtcp_sdes_t *) malloc(sizeof(rtcp_sdes_t))) == NULL) {
    perror("rtcp_packet_alloc() malloc() #4");
    exit(-2);
  }

  if ((p->sdes->chunk_item_cname = (char *) malloc(sizeof(char) * RTCP_MAX_SDES_ITEM_LENGTH+8)) == NULL) {
    perror("rtcp_packet_alloc() malloc() #5");
    exit(-2);
  }

  if ((p->fb = (rtcp_fb_t *) malloc(sizeof(rtcp_fb_t))) == NULL) {
    perror("rtcp_packet_alloc() malloc() #6");
    exit(-2);
  }

  p->fb->n_fci = 0;
  
  if ((p->packet = (byte *) malloc(sizeof(byte)*RTCP_MAX_PACKET_SIZE)) == NULL) {
    perror("rtcp_packet_alloc() malloc() #6");
    exit(-2);
  }
  
  *pp = p;
}

/*
 * Free a RTCP packet from memory
 * */
static void rtcp_packet_free(rtcp_packet_t *p)
{
  assert(p);

  if (p->rr->rb) {free(p->rr->rb); p->rr->rb = NULL;}
  if (p->rr) {free(p->rr); p->rr = NULL;}
  if (p->sdes->chunk_item_cname) {free(p->sdes->chunk_item_cname); p->sdes->chunk_item_cname = NULL;}
  if (p->sdes) {free(p->sdes); p->sdes = NULL;}
  if (p->fb) {
    if (p->fb->fci) {free(p->fb->fci); p->fb->fci = NULL;}
    free(p->fb); p->fb = NULL;
  }
  if (p->packet) {free(p->packet); p->packet = NULL;}
}

/*
 * Reset the values in the allocated RTCP packet
 * */
static void rtcp_packet_reset(rtcp_packet_t *p)
{
  assert(p);
  assert(p->rr);
  assert(p->sdes);
  assert(p->sdes->chunk_item_cname);
  assert(p->fb);

  memset(p->sdes->chunk_item_cname, 0, RTCP_MAX_SDES_ITEM_LENGTH);

  /* We keep the same allocated memory except for FCI messages */
  if (p->fb->fci) {free(p->fb->fci); p->fb->fci = NULL;}
  p->fb->n_fci = 0;

  /* Do nothing more, just remember to assign new values before composing
   * packet */
}

/*
 * Add an FB message to the RTCP packet
 * */
static void rtcp_add_fb (uint16_t lostseq)
{
  assert(rtcp_packet->fb);
  
  rtcp_fci_nack_t *fci;
  int c, i;

  // todo: (low pri) add support for cycling seq
  
  c = rtcp_packet->fb->n_fci;
  /* Check if info can be added to existing FB message */
  if (c > 0) {
    for (i=c-1; i>=0; i--) {
      fci = &rtcp_packet->fb->fci[i];
      if (lostseq <= fci->pid + 16 ) {
        fci->blp |= (0x01 << (lostseq - fci->pid - 1));
        RTCP_DEBUG(("Altered BLP of FCI[%d], pid=%d, lostseq=%d, blp=0x%04X\n", i, fci->pid, lostseq, fci->blp));
        return;
      }  
    }
  }

  /* Can't add to existing FBs, create a new one */
  if (c == 0) {
    /* First FCI. Allocate new area. */
    if ((rtcp_packet->fb->fci = (rtcp_fci_nack_t *) malloc(sizeof(rtcp_fci_nack_t))) == NULL) {
      perror("Error allocating memory for FCI\n");
      return;
    }
  }
  else {
    /* There exist FCIs from before, extend allocated area */
    if ((rtcp_packet->fb->fci =
         (rtcp_fci_nack_t *) realloc(rtcp_packet->fb->fci, sizeof(rtcp_fci_nack_t) * c+1)) == NULL)
      perror("Error reallocating memory for FCI\n");
  }

  rtcp_packet->fb->fci[c].pid = lostseq;
  rtcp_packet->fb->fci[c].blp = 0;
  rtcp_packet->fb->n_fci += 1;
  RTCP_DEBUG(("New FCI added [%d], lostseq=%d\n", c, lostseq));
}


/*
 * Add a new FB message to the RTCP packet
 * */
static void rtcp_new_fb(uint16_t lostseq, double t0)
{
//  double te;
  double const t_dither_max = RTCP_T_DITHER_MAX;

  rtcp_add_fb(lostseq);
  
  te = t0 + drand48() * t_dither_max;
//  rtcp_schedule_early_fb(te);
}


/* Static: Send RTCP RR */
static int rtcp_send_rr (int id)
{
  int pck_size;
  static int rrno = 1;

  rtcp_set_rr(rtcp_packet);
  rtcp_set_sdes(rtcp_packet);
  rtcp_set_fb(rtcp_packet);
  pck_size = rtcp_compose_packet(rtcp_packet);

  RTCP_PRINT_EVENT(("Sending RR # %d", rrno++), RTCP_SEND_RR);
//  printf("%.3f: Sending RR # %d\n", GET_TIME_SHORT(timenow()), rrno++);
  RTCP_DEBUG(("Sending RR id=%d\n", id));
  if (rtcp_send_packet(rtcp_packet, &rtcp_con) < 0)
    perror("Error sending regular RTCP, ignoring\n");
  rtcp_packet_reset(rtcp_packet);

  return pck_size;
}


/*
 * Used for creating a byte representation of the structure rtcp_packet, which
 * is placed into its field rtcp_packet->packet. It is assumed that the
 * receiver report, the sdes and feedback message(s) referred to in the
 * rtcp_packet structure already holds the correct values.
 * 
 * The return value is the size of this packet in bytes, which may also be
 * found in rtcp_packet->length after return.
 **/
static int rtcp_compose_packet(rtcp_packet_t *p)
{
  int i;
  uint16_t temp16;
  uint32_t temp32;

  assert(p);
  assert(p->packet);

  p->packlen = 0;
    
  /* The RR must be first, then the SDES, and then all FB messages  */

  /* RR */
  /* The following code are only valid for exactly one RR */
  assert(p->rr->rc == 1);
  assert(p->rr->length == 7);
  
  p->packet[0] = ( ((p->rr->v & 0x03) << 6)
                   | ((p->rr->p & 0x01) << 5)
                   | ((p->rr->rc & 0x1F) << 0));

  p->packet[1] = (byte) p->rr->pt;

  temp16 = htons(p->rr->length);
  memcpy(&p->packet[2], &temp16, 2);

  temp32 = htonl(p->rr->ssrc);
  memcpy(&p->packet[4], &temp32, 4);

  temp32 = htonl(p->rr->rb->ssrc_n);
  memcpy(&p->packet[8], &temp32, 4);

  p->packet[12] = (byte) p->rr->rb->fl;

  temp32 = htonl(p->rr->rb->cnlp << 8); /* want only 3 bytes on from */
  memcpy(&p->packet[13], &temp32, 3);

  temp32 = htonl(p->rr->rb->ehsn);
  memcpy(&p->packet[16], &temp32, 4);
  
  temp32 = htonl(p->rr->rb->jitter);
  memcpy(&p->packet[20], &temp32, 4);

  temp32 = htonl(p->rr->rb->lsr);
  memcpy(&p->packet[24], &temp32, 4);

  temp32 = htonl(p->rr->rb->dlsr);
  memcpy(&p->packet[28], &temp32, 4);

  p->packlen = 32;

  /* SDES (mandatory according to RFC4585) */
  /* The following are only valid of you have exactly one chunk with only
   * cname */
  assert(p->sdes->sc == 1);
  assert(p->sdes->chunk_item_type == 1);
  
  p->packet[32] = ( ((p->sdes->v & 0x03) << 6)
                   | ((p->sdes->p & 0x01) << 5)
                   | ((p->sdes->sc & 0x1F) << 0));

  p->packet[33] = (byte) p->sdes->pt;

  temp16 = htons(p->sdes->length);
  memcpy(&p->packet[34], &temp16, 2);

  p->packlen = 36;

  temp32 = htonl(p->sdes->chunk_csrc);
  memcpy(&p->packet[36], &temp32, 4);

  p->packet[40] = p->sdes->chunk_item_type;
  p->packet[41] = p->sdes->chunk_item_length;

  /* todo: (low pri) make this network byte order */
  memcpy(&p->packet[42], p->sdes->chunk_item_cname, p->sdes->chunk_item_length);
  p->packlen = 42 + p->sdes->chunk_item_length;

  /* Chunk should be ended with a zero byte */
  memset(&p->packet[p->packlen], 0, 1);
  p->packlen += 1;
  /* And zero pad until next 32-bit word */

  memset(&p->packet[p->packlen], 0, (4 - p->packlen % 4) % 4);
  p->packlen += (4 - p->packlen % 4) % 4;

  
  /* FB */
  /* If there's no FCI don't append FB message */
  if (p->fb->n_fci == 0)
    return p->packlen;
  
  p->packet[p->packlen] = ( ((p->fb->v & 0x03) << 6)
                   | ((p->fb->p & 0x01) << 5)
                   | ((p->fb->fmt & 0x1F) << 0));

  p->packet[p->packlen+1] = (byte) p->fb->pt;

  temp16 = htons(p->fb->length);
  memcpy(&p->packet[p->packlen+2], &temp16, 2);

  temp32 = htonl(p->fb->ssrc_sender);
  memcpy(&p->packet[p->packlen+4], &temp32, 4);

  temp32 = htonl(p->fb->ssrc_source);
  memcpy(&p->packet[p->packlen+8], &temp32, 4);

  p->packlen += 12;

  for (i = 0; i < p->fb->n_fci; i++) {
    /* Ignore rest of fb's max packet size is reached */
    if (p->packlen + 4 > RTCP_MAX_PACKET_SIZE)
      break;
    
    temp16 = htons(p->fb->fci[i].pid);
    memcpy(&p->packet[p->packlen], &temp16, 2);

    temp16 = htons(p->fb->fci[i].blp);
    memcpy(&p->packet[p->packlen + 2], &temp16, 2);

    p->packlen += 4;
  }

  return p->packlen;
}


/*
 * Calculate receiver statistics based on information in the structure
 * rtcp_stats and set all the fields of the rr packet in the rtcp_packet
 * structure.
 * */
static void rtcp_set_rr(rtcp_packet_t *p)
{
  assert(p);
  
  const unsigned short v = 2;
  const unsigned short pad = 0;                     
  const unsigned short rc = 1;          /* We always have only one report block */
  const unsigned short pt = RTCP_RR;                    
  const unsigned short length = 7;      /* # 32-bit words minus 1. Always 7
                                             since we have only one report
                                             block. */
  const unsigned long ssrc = RTCP_SSRC_DECODER;
  const unsigned long ssrc_1 = RTCP_SSRC_ENCODER;
  
  uint32_t extended_max;
  uint32_t expected;                      /* total expected received packets */
  uint32_t expected_interval;             /* expected received since last rr */
  uint32_t received_interval;             /* acutally received since last rr */
  uint32_t lost;                          /* lost packets in total */
  uint32_t lost_interval;                 /* lost packets since last rr */
  uint8_t fraction;

  /* Calculate and update stats according to A.3 RFC 3550 */
  extended_max = rtcp_stats.cycles + rtcp_stats.max_seq;
  expected = extended_max - rtcp_stats.base_seq + 1;

  lost = expected - rtcp_stats.received; 

  expected_interval = expected - rtcp_stats.expected_prior;
  rtcp_stats.expected_prior = expected;
  received_interval = rtcp_stats.received - rtcp_stats.received_prior;
  rtcp_stats.received_prior = rtcp_stats.received;
  lost_interval = expected_interval - received_interval;

  if (expected_interval == 0 || lost_interval <= 0)
    fraction = 0;
  else
    fraction = (lost_interval << 8) / expected_interval;

  /* Set the values into the RR struct */
  p->rr->v = v;
  p->rr->p = pad;
  p->rr->rc = rc;
  p->rr->pt = pt;
  p->rr->length = length;
  p->rr->ssrc = ssrc;

  p->rr->rb->ssrc_n = ssrc_1;
  p->rr->rb->fl = fraction;
  p->rr->rb->cnlp = lost;
  p->rr->rb->ehsn = extended_max;

  /* Ignore the jitter and delay stuff since we're not sending any SRs */
  p->rr->rb->jitter = 0;
  p->rr->rb->lsr = 0;
  p->rr->rb->dlsr = 0;
}

/*
 * Set the SDES of the rtcp_packet structure.
 * */
static void rtcp_set_sdes(rtcp_packet_t *p)
{
  assert(p);
  
  
  const uint8_t v = 2;
  const uint8_t pad = 0;
  const uint8_t sc = 1;
  const uint8_t pt = RTCP_SDES;
  const unsigned long chunk_csrc = RTCP_SSRC_DECODER;
  const unsigned short chunk_item_type = 1; /* CNAME */
  const char *chunk_item_cname = "stian@master.ntnu.no";
  const unsigned short chunk_item_cname_length = strlen(chunk_item_cname);
  assert(chunk_item_cname_length <= RTCP_MAX_SDES_ITEM_LENGTH);

  const uint8_t length = (chunk_item_cname_length + 2)/4 + 2;        /* # 32-bit words minus 1 */

  
  p->sdes->v = v;
  p->sdes->p = pad;
  p->sdes->sc = sc;
  p->sdes->pt = pt;
  p->sdes->length = length;

  p->sdes->chunk_csrc = chunk_csrc;
  p->sdes->chunk_item_length = chunk_item_type;
  p->sdes->chunk_item_type = chunk_item_type;
  p->sdes->chunk_item_length = chunk_item_cname_length;
  strcpy(p->sdes->chunk_item_cname, chunk_item_cname);
}

/*
 * Set the FB message (only headers of the FB message)
 * */
static void rtcp_set_fb(rtcp_packet_t *p)
{
  assert(p);
  
  const uint8_t v = 2;
  const uint8_t pad = 0;
  const uint8_t fmt = 1;                /* Generick NACK */
  const uint8_t pt = RTCP_RTPFB;               /* Transport layer FB message */
  const uint16_t length = 2 + p->fb->n_fci;    /* because of GNACK is 32 bit
                                                  long */

  const uint32_t ssrc_sender = RTCP_SSRC_DECODER;
  const uint32_t ssrc_source = RTCP_SSRC_ENCODER;

  p->fb->v = v;
  p->fb->p = pad;
  p->fb->fmt = fmt;
  p->fb->pt = pt;
  p->fb->length = length;
  p->fb->ssrc_sender = ssrc_sender;
  p->fb->ssrc_source = ssrc_source;

  /* The FCI is already set since FB messages are added continously */
}

/*
 * Send the RTCP packet to given connection. Returns a negative value if not
 * successful.
 * */
static int rtcp_send_packet(rtcp_packet_t *p, rtcp_connection_t *con)
{
  assert(p);
  assert(con);
  
  if (p->packlen != sendto(con->sockfd_out, p->packet, p->packlen, 0,
                           (struct sockaddr *) &con->encoder_addr,
                           sizeof(con->encoder_addr))) {
    perror("Error sending RTCP packet");
    return -1;
  }
  
  return 0;
}

/*
 * Send RTCP FB
 * */
static void rtcp_send_early_fb (void)
{
  int pck_size;
  double t_rr;
  double tn_old;
  static int fbno = 1;

  
  /* Sending an early FB means sending a compound RTCP packet, that is a RTCP
   * packet that includes both one (or more) RR, SDES and FB message. We only
   * use minimal compound messages (exactly one RR, SDES and FB). */
  rtcp_set_rr(rtcp_packet);
  rtcp_set_sdes(rtcp_packet);
  rtcp_set_fb(rtcp_packet);
  pck_size = rtcp_compose_packet(rtcp_packet);

  RTCP_PRINT_EVENT(("Sending early FB # %d", fbno++), RTCP_SEND_FB);
  if (rtcp_send_packet(rtcp_packet, &rtcp_con) < 0)
    fprintf(stderr, "Error sending early feedback, ignoring\n");
  
  rtcp_packet_reset(rtcp_packet);

  tn_old = tn;
  t_rr = rtcp_interval();
  
  if (t_rr_last == TIME_NaN)
    /* This is not in the standard, but some alternative calculation of tn is
     * needed as tp still is 0 (its initial value) if no RR has been sent. */
    tn = timenow() + t_rr;
  else
    tn = tp + 2 * t_rr;

  tp = tn_old;

  allow_early = 0;
  rtcp_early_fb_waiting = 0;
  te = TIME_NO_EARLY_FB;
  avg_rtcp_size = 1./16 * pck_size + 15./16 * avg_rtcp_size;
  

  /* According to 3.5.2.6:
   *
   * allow_early = false
   *
   * recalc tn (next regular RTCP) and set tp:
   * tn = tp + 2*T_rr (means send fb instead of the next regular packet)
   * 
   *   
   * Schedule allow_early = true for time tn
   *
   * */
  
}


/*
 * Send RTCP RR message maybe
 * */
static void rtcp_send_rr_maybe(void)
{
  /* Only for P2P. Session members (pmembers) are ignored */
  
  double t_rr;                          /* interval */
  double tc;                            /* current time */
  int pck_size;                         /* sent packet size */
  double t_rr_current_interval;         
  static int rrid = 1;

  tc = timenow();
  
  
  /* todo: remove all timing reconsideration? */
  
  /* No minimal interval between regular RTCP packets */
  if (t_rr_interval == 0) {

    /* Send and update variables. Will definitely go into here if first call
     * since tp is initiated as 0. */
#if RTCP_ENABLE_TIMER_RECONSIDERATION
    t_rr = rtcp_interval();
    if (tp + t_rr <= tc) {
      RTCP_DEBUG(("(1) Send RR %d, tn = %.3f, tp+t_rr=%.3f\n", rrid, tn, tp+t_rr));
#else
      RTCP_DEBUG(("(1) Send RR %d, tn = %.3f\n", rrid, tn));
#endif

      pck_size = rtcp_send_rr(rrid);
      avg_rtcp_size = 1./16 * pck_size + 15./16 * avg_rtcp_size;
      tp = tc;
      initial = 0;
      
      /* calculate new interval and schedule */
      t_rr = rtcp_interval();
      tn = tc + t_rr;
      allow_early = 1;
      rtcp_rr_fb_waiting = 0;
#if RTCP_ENABLE_TIMER_RECONSIDERATION
    }
    /* Reschedule */
    else {
      RTCP_DEBUG(("(2) reschedule RR %d, timing reconsideration, tn = %.3f, tp = %.3f, t = %.3f\n", tc, rrid, tn, tp, t_rr));
      tn = tp + t_rr;
    }
#endif
  }
  
  /* We must consider the minimum interval between regular RTCP packets */
  else {
    /* Timing reconsideration. Will definitely go into here if first call
     * since tp is initiated 0. */
#if RTCP_ENABLE_TIMER_RECONSIDERATION
    t_rr = rtcp_interval();
    if (tp + t_rr <= tc) {
#endif
      /* If no RR sent before we must send or schedule one */
      if (t_rr_last == TIME_NaN) {
        RTCP_DEBUG(("(3) Send RR %d, no last rr, tn = %.3f, t_rr_last = %f\n", rrid, tn, t_rr_last));
        /* send and update */
        pck_size = rtcp_send_rr(rrid);
        avg_rtcp_size = 1./16 * pck_size + 15./16 * avg_rtcp_size;
        t_rr_last = tn;
      }

      /* Compare against "random" minimal interval and decide */
      else {
        t_rr_current_interval = (drand48() + 0.5) * t_rr_interval;

        /* tn is over "random" minimal interval */
        if (t_rr_last + t_rr_current_interval <= tn) {
          RTCP_DEBUG(("(4) Send RR %d, after min interval, tn = %.3f, t_rr_last = %.3f, t_rr_current_interval = %.3f\n", rrid, tn, t_rr_last, t_rr_current_interval));
          pck_size = rtcp_send_rr(rrid);
          avg_rtcp_size = 1./16 * pck_size + 15./16 * avg_rtcp_size;
          t_rr_last = tn;
        }
      
        /* Under the "random" minimum interval, but send anyway if FB messages
         * are waiting */
        else if (rtcp_rr_fb_waiting && t_rr_last + t_rr_current_interval > tn ) {
          RTCP_DEBUG(("(5) Send RR %d, too early but fb, tn = %.3f, t_rr_last = %.3f, t_rr_current_interval = %.3f\n", rrid, tn, t_rr_last, t_rr_current_interval));
          pck_size = rtcp_send_rr(rrid);
          avg_rtcp_size = 1./16 * pck_size + 15./16 * avg_rtcp_size;
        }
        else {
          RTCP_DEBUG(("(6) Suppress RR %d, too early, tn = %.3f, t_rr_last = %.3f, t_rr_current_interval = %.3f\n", rrid, tn, t_rr_last, t_rr_current_interval));
          // Suppress packet
        }
      }
      // Schedule the next rr
      t_rr = rtcp_interval();
      tn = tc + t_rr;

      tp = tc;
      allow_early = 1;
      rtcp_early_fb_waiting = 0;
      rtcp_rr_fb_waiting = 0;

      initial = 0;
#if RTCP_ENABLE_TIMER_RECONSIDERATION
    } /* if (tn <= tc) */

    else {
      RTCP_DEBUG(("(7) Reschedule RR %d due to timing reconsideration, tn = %.3f, tp = %.3f, t = %.3f\n", tc, arg->id, tn, tp, t_rr));
      tn = tp + t_rr;
    }
#endif
  }

  rrid++;
}

/*
 * Send RTCP FB message maybe
 * */
static void rtcp_send_early_fb_maybe(uint16_t seq)
{
  double t0;
  double const t_dither_max = RTCP_T_DITHER_MAX;
  double const t_max_fb_delay = RTCP_T_MAX_FB_DELAY;

  pthread_mutex_lock(&rtcp_mutex);

  /* FB message(s) already schedule */
  if (rtcp_early_fb_waiting) {
    RTCP_PRINT_EVENT(("FB message(s) is scheduled, addding FB to FB"), RTCP_ADD_FB_TO_FB);
    rtcp_add_fb(seq);
    goto finished;
  }
  else if (rtcp_rr_fb_waiting) {
    RTCP_PRINT_EVENT(("FB message(s) is scheduled, addding FB to RR"), RTCP_ADD_FB_TO_RR);
    rtcp_add_fb(seq);
    goto finished;
  }

  /* First FB since last RTPC packet */
  else {
    t0 = timenow();

    if (t0 + t_dither_max > tn) {
      RTCP_DEBUG(("Sending RR soon, adding FB (t0 = %.3f, d_max = %.3f, tn = %.3f)\n", t0, t_dither_max, tn));
      RTCP_PRINT_EVENT(("Sending RR soon, adding FB to RR"), RTCP_ADD_FB_TO_RR);
      rtcp_add_fb(seq);
      rtcp_rr_fb_waiting = 1;
    }
    else {
      if (allow_early) {
        //te = t0 + drand48() * t_dither_max;
        RTCP_DEBUG(("Allowed to send early FB\n"));
        RTCP_PRINT_EVENT(("Early FB is scheduled"), RTCP_SCHEDULE_FB);
        rtcp_new_fb(seq, t0);
        rtcp_early_fb_waiting = 1;
        //rtcp_schedule_early_fb(te);
      }
      else {
        if (tn - t0 < t_max_fb_delay) {
          RTCP_PRINT_EVENT(("Early FB not allowed, addding FB to RR"), RTCP_ADD_FB_TO_RR);
          //printf("%.3f: Early FB not allowed, but adding to RR\n", GET_TIME_SHORT(timenow()));
          rtcp_add_fb(seq);
          rtcp_rr_fb_waiting = 1;
        }
        else {
          RTCP_PRINT_EVENT(("Early FB not allowed, discarded"), RTCP_LOSS_IGNORED);
          //printf("%.3f: Early FB not allowed, discarded\n", GET_TIME_SHORT(timenow()));
          // discard
        }
      }
    }

    /* Detection of overlap between FB messages from other receivers are
     * skipped because it's not relevant for a p2p application */
  }
  
  
  /* according to 3.5.2:
   *
   * called when packet(s) is lost an fb is useful
   *
   * Check if there is a RTCP packet in scheduled for transmission
   * (one variable that is set when scheduled and reset when sent)
   *
   * If so, add this FB to that packet (rtcp_add_fb()). Finished
   *
   * If not, create a new compound (min or full) RTCP packet with T_dither_max
   * = 0 because p2p (Larger value is allowed considering application
   * specific needs) (T_dither_max is max additional delay and may be needed
   * for reporting bursts)
   *
   * If regular RTCP packet is to be sent before t0 + T_dither_max, add this
   * FB to that packet. Finished.
   *
   * If not allow_early:
   *    If tn - t0 < T_max_fb_delay
   *        Add FB to regular packet
   *    Else
   *        Discard FB
   * Else
   *    Schedule early fb on te = t0 + RND[0,1] * T_dither_max (rtcp_send_early_fb())
   *
   * 
   * */

 finished:
  pthread_mutex_unlock(&rtcp_mutex);
}

/*
 * Calulate the transmission interval
 * */
static double rtcp_interval()
{
  /* Minimum average time between RTCP packets from this site (in   */
  /* seconds).  This time prevents the reports from `clumping' when */
  /* sessions are small and the law of large numbers isn't helping  */
  /* to smooth out the traffic.  It also keeps the report interval  */
  /* from becoming ridiculously small during transient outages like */
  /* a network partition.                                           */
//	double const RTCP_MIN_TIME = 5.0;
  /* Fraction of the RTCP bandwidth to be shared among active       */
  /* senders.  (This fraction was chosen so that in a typical       */
  /* session with one or two active senders, the computed report    */
  /* time would be roughly equal to the minimum report time so that */
  /* we don't unnecessarily slow down receiver reports.) The        */
  /* receiver fraction must be 1 - the sender fraction.             */
//	double const RTCP_SENDER_BW_FRACTION = 0.25;
//	double const RTCP_RCVR_BW_FRACTION   = (1-RTCP_SENDER_BW_FRACTION);
  /* To compensate for "unconditional reconsideration" converging   */
  /* to a value below the intended average.                         */
  double const COMPENSATION            = 2.71828 - 1.5;

  double t;				              /* interval */
//	double rtcp_min_time = RTCP_MIN_TIME;
  int n;			        /* no. of members for computation */

//  double rtcp_bw = RTCP_BW;
//  double rtcp_bw = input->rtp_bw * RTCP_BW_FRACTION;
  double rtcp_bw = input->rtp_bw / 8.0 * 1000 * RTCP_BW_FRACTION; /* bytes/sec */
  double rtcp_min_time = RTCP_MIN_TIME;

  /* Very first call at application start-up uses half the min      */
  /* delay for quicker notification while still allowing some time  */
  /* before reporting for randomization and to learn about other    */
  /* sources so the report interval will converge to the correct    */
  /* interval more quickly.                                         */
  if (initial) {
    rtcp_min_time /= 2;
  }

  /*
   * Dedicate a fraction of the RTCP bandwidth to senders unless
   * the number of senders is large enough that their share is
   * more than that fraction.
   */
  n = members;
  if (senders <= members * RTCP_SENDER_BW_FRACTION) {
    if (we_sent) {
      rtcp_bw *= RTCP_SENDER_BW_FRACTION;
      n = senders;
    } else {
      rtcp_bw *= RTCP_RCVR_BW_FRACTION;
      n -= senders;
    }
  }

  /*
   * The effective number of sites times the average packet size is
   * the total number of octets sent when each site sends a report.
   * Dividing this by the effective bandwidth gives the time
   * interval over which those packets must be sent in order to
   * meet the bandwidth target, with a minimum enforced.  In that
   * time interval we send one report so this time is also our
   * average time between reports.
   */
  t = avg_rtcp_size * n / rtcp_bw;
  if (t < rtcp_min_time) t = rtcp_min_time;

  /*
   * To avoid traffic bursts from unintended synchronization with
   * other sites, we then pick our actual next report interval as a
   * random number uniformly distributed between 0.5*t and 1.5*t.
   */
  t = t * (drand48() + 0.5);
  t = t / COMPENSATION;
  return t;
}

/*
 * Init the RTCP connection to the sender
 * */
void rtcp_init_connection()
{
  strcpy(rtcp_con.encoder_host, input->rtcpHost);
  rtcp_con.encoder_port = input->rtcpHostPort;

  memset(&rtcp_con.encoder_addr, 0, sizeof(rtcp_con.encoder_addr));
  rtcp_con.encoder_addr.sin_family = AF_INET;
  inet_pton(AF_INET, rtcp_con.encoder_host, &rtcp_con.encoder_addr.sin_addr);
  rtcp_con.encoder_addr.sin_port = htons(rtcp_con.encoder_port);

  if ((rtcp_con.sockfd_out = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("Failed connecting RTCP socket, exit\n");
    exit(-1);
  }
}

/*
 * Initialize statistics
 * */
void rtcp_init_stats (uint16_t base_seq)
{
  rtcp_stats.base_seq = base_seq;
  rtcp_stats.max_seq = base_seq;
  rtcp_stats.cycles = 0;
  rtcp_stats.received = 0;
  rtcp_stats.expected_prior = 0;
  rtcp_stats.received_prior = 0;  
}

/*
 * The main loop for sending Early FB messages
 * */
static void *rtcp_fb_loop (void *arg)
{
  const int loop_interval = 5000;              /* usec */
  do {
    pthread_mutex_lock(&rtcp_mutex);

    if (rtcp_early_fb_waiting && te < timenow())
      rtcp_send_early_fb();

    pthread_mutex_unlock(&rtcp_mutex);

    usleep(loop_interval);
  } while (rtcp_alive);

  pthread_exit(NULL);
}


/*
 * The main loop for sending regular RTCP packets
 * */
static void *rtcp_rr_loop (void *arg)
{
  do {
    pthread_mutex_lock(&rtcp_mutex);
    
    if (tn <= timenow()) 
      rtcp_send_rr_maybe();
    else {
      RTCP_DEBUG(("(0) RR ignored. tn=%.3f now=%.3f\n", tn, timenow()));
    }
      

    pthread_mutex_unlock(&rtcp_mutex);
    
    /* tn has been set in rtcp_send_rr_maybe */
    usleep((tn - timenow()) * 1000000);
  } while (rtcp_alive);

  pthread_exit(NULL);
}


/*
 * Initialize the RTCP functionality and variables
 * */
void rtcp_init(uint16_t base_seq)
{
  /* init variables according to sec 3.5.1
   *
   * set T_rr_interval
   * allow_early
   * t_rr_last = NaN
   *
   * Init vars as RFC 3550 sec 6.3.2 except Tmin
   *
   * Tmin = 0 (dependent on T_rr_interval?)
   *
   * Calculate next exp time Tn and schedule
   *
   * 
   * */

/* This should be the initial values and updated in runtime. But since we know
  our values and they're constant we set them initially.
  senders = 0;
  members = 1;
  pmembers = 1;
*/
  senders = 1;
  members = 2;
  pmembers = 2;

  we_sent = 0;
  initial = 1;
  allow_early = 1;

  avg_rtcp_size = (double) RTCP_PROBABLE_RTCP_SIZE;

  tp = 0;
  t_rr_last = TIME_NaN;
  t_rr_interval = RTCP_T_RR_INTERVAL;

  next_rr_id = 0;
  next_fb_id = 0;
  rtcp_early_fb_waiting = 0;
  rtcp_rr_fb_waiting = 0;
  te = TIME_NO_EARLY_FB;

  rtcp_init_stats(base_seq);
  rtcp_init_connection();

  rtcp_packet_alloc(&rtcp_packet);

  tn =  timenow() + rtcp_interval();
  rtcp_alive = 1;

  /* Start loop for regular RTCP packet in separate thread */
  pthread_t thr;
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  if (pthread_create(&thr, NULL, rtcp_rr_loop, NULL) != 0)
    perror("Failed creating RR thread");
  
  /* Start loop for FB messages in a new thread */ 
  if (pthread_create(&thr, NULL, rtcp_fb_loop, NULL) != 0)
    perror("Failed creating FB thread");
  
  pthread_attr_destroy(&attr);

}

/* ******************************************************************************** */
/* TEST CODE FOR RTCP */
/* ******************************************************************************** */
#ifdef RTCP_TEST
#include <time.h>
#include <unistd.h>
#include <pthread.h>

typedef struct {
  int i;
  double t_called;
  double t_interval;
} printNow_t;


void printNow(void *arg)
{
  printNow_t *args = (printNow_t *) arg;
  printf("Callback no %d\t called at %.3f\t dealy %.3f\t timenow %.3f\n",
         args->i, args->t_called, args->t_interval, timenow());
}


void testScheduleRR()
{
  int i;
  
  for (i = 0; i < 10; i++) {
    rtcp_schedule_rr(timenow() + rtcp_interval());
    sleep(5);
  }

  pthread_exit(NULL);
  
}

void testSchedule()
{
  int i;
  
  for (i = 0; i < 10; i++) {
    schedule_t *s = (schedule_t *) malloc(sizeof(schedule_t));
    printNow_t *a = (printNow_t *) malloc(sizeof(printNow_t));
    
    a->i = i;
    a->t_called = timenow();
    a->t_interval = rtcp_interval();

    s->usec = a->t_interval * 1000000;
    s->callback = printNow;
    s->args = a;

    schedule(s);
    
    sleep(1);
  }

  pthread_exit(NULL);
}


void testTimenow()
{
  int i;

  for (i = 0; i < 10; i++) {
    printf("Seconds since epoch: %.3f\n", timenow());
    sleep(1);
  }
}

void testInterval()
{
  double t;
  int i;

  for (i = 0; i < 10; i++) {
    t = rtcp_interval();
    printf("Calculated interval no %d: %.3f\n", i, t);
  }
}

void testFB()
{
  double const q = 0.3;
  double const p = 0.02;
  int packet_interval = 10000;
  unsigned int j = 1;
  RTPpacket_t pck = {0,0,0,0,0,0,0,0,0,0,NULL,0,NULL,0};

  rtcp_init(0);
  pck.seq = 0;
  rtcp_rtp_packet_received(&pck);


  for(;;j++)
  {
    if (drand48() < p) {
      RTCP_DEBUG(("Lost packet %d\n", j));
      rtcp_send_early_fb_maybe(j);
      j++;
      while (drand48() < q) {
        RTCP_DEBUG(("Lost consecutive packet %d\n", j));
        usleep(packet_interval);
        rtcp_send_early_fb_maybe(j);
        j++;
      }
    }
    else {
      pck.seq = j;
      rtcp_rtp_packet_received(&pck);
    }
    usleep(packet_interval);
  }
}

int main(int argc, char **argv)
{

//  srand48(time(NULL));
  
  testFB();

//  testScheduleRR();
//  exit(0);
  pthread_exit(NULL);
}

#endif


#endif /* RTCP_FEEDBACK */
