
/*!
 ************************************************************************
 * \file  rtp.c
 *
 * \brief
 *    Network Adaptation layer for RTP packets
 *
 * \author
 *    Main contributors (see contributors.h for copyright, address and affiliation details)
 *    - Stephan Wenger   <stewe@cs.tu-berlin.de>
 ************************************************************************
 */


/*!

  A quick guide to the basics of the RTP decoder implementation

  This module contains the RTP packetization, de-packetization, and the
  handling of Parameter Sets, see VCEG-N52 and accompanying documents.
Note: Compound packets are not yet implemented!

The interface between every NAL (including the RTP NAL) and the VCL is
based on Slices.  The slice data structure on which the VCL is working
is defined in the type Slice (in defines.h).  This type contains the
various fields of the slice header and a partition array, which itself
contains the data partitions the slice consists of.  When data
partitioning is not used, then the whole slice bit string is stored
in partition #0.  When individual partitions are missing, this is
indicated by the size of the bit strings in the partition array.
A complete missing slice (e.g. if a Full Slice packet was lost) is
indicated in a similar way.  

part of the slice structure is the error indication (ei-flag).  The
Ei-flag is set in such cases in which at least one partition of a slice
is damaged or missing.When data partitioning is used, it can happen that
one partition does not contain any symbols but the ei_flag is cleared,
which indicates the intentional missing of symbols of that partition.
A typical example for this behaviour is the Intra Slice, which does not
have symnbols in its type C partition.

The VCL requests new data to work on through the call of readSliceRTP().
This function calls the main state machine of this module in ReadRTPpaacket().

ReadRTPpacket assumes, when called, that in an error free environment
a complete slice, consisting of one Full Slice RTP packet, or three Partition
packets of types A, B, C with consecutive sequence numbers, can be read.
It first interprets any trailing SUPP and Parameter Update (Header) packets.
Then it reads one video data packet.  Two cases have to be distinguished:

1. Type A, or Full Slice packet
In this case, the PictureID and the macroblock mumbers are used to
identify the potential loss of a slice.  A slice is lost, when the
StartMB of the newly read slice header is not equal to the current
state of the decoder
1.1 Loss detected
In this case the last packet is unread (fseek back), and a dummy slice
containing the missing macroblocks is conveyed to the VCL.  At the next 
call of the NAL, the same packet is read again, but this time no packet 
loss is detected by the above algorithm,
1.2. No loss
In this case it is checked whether a Full Slice packet or a type A data
partition was read
1.2.1 Full Slice
The Full Slice packet is conveyed to the NAL
1.2.2 Type A Partition
The function RTPReadDataPartitionedSlice() is called, which collects
the remaining type B, C partitions and handles them appropriately.

Paraneter Update Packets (aka Header packets) are in an SDP-like syntax
and are interpreted by a simple parser in the function 
RTPInterpretParameterSetPacket() 

Each Slice header contaions the information on which parameter set to be used.
The function RTPSetImgInp() copies the information of the relevant parameter
set in the VCL's global variables img-> and inp->  IMPORTANT: any changes
in the semantics of the img-> and inp-> structure members must be represented
in this function as well!

A note to the stream-buffer data structure: The stream buffer always contains
only the contents of the partition in question, and not the slice/partition
header.  Decoding has to start at bitoffset 0 (UVLC) or bytreoffset 0 (CABAC).

The remaining functions should be self-explanatory.

*/

#include "contributors.h"

#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "global.h"
#include "errorconcealment.h"
#include "rtp.h"
#include "fmo.h"
#include "sei.h"
#include "memalloc.h"
#include "defines.h"

#ifdef WIN32
#include <Winsock2.h>
#else
#include <netinet/in.h>
#endif

#if RTP_BIG_ENDIAN_SUPPORT
#include "endianness.h"
#endif

FILE *bits;

#if RTP_OVER_NETWORK

#include <sys/timeb.h>
#include <limits.h>
#include <sys/select.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "rtcp.h"

#define MAX_RTPPACKETS_IN 32   /* The number of max packets being reordered */
 /* Next timestamp when cache is empty.  */
#define RTP_CACHE_EMPTY_TIMESTAMP ULONG_MAX - 100L*RTP_MAX_DELAY
#define RTP_MAX_DELAY 50 /* milliseconds */

#ifndef MIN
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
#endif

#define RTP_DEBUG_ENABLE 0

#if RTP_DEBUG_ENABLE
#define RTP_DEBUG(s) printf s
#else
#define RTP_DEBUG(s)
#endif



static RTPConnection con;
static struct timeb timeStart;
static unsigned long firstTimestamp;
static int p_received;

typedef struct {
  RTPpacket_t  *pack[MAX_RTPPACKETS_IN]; /* Pointers to the RTP packets cached */
  unsigned short seq[MAX_RTPPACKETS_IN]; /* Seq no for the RTP packet */
  unsigned short len[MAX_RTPPACKETS_IN]; /* Length of packet. len[i] > 0 if
                                          * there is at packet at pos i. */
  unsigned short first;                  /* Index of next sequence slot of
                                          * expected packet */
  unsigned int nextTimestamp;                    /* The next timestamp expiring in buffer */
} rtpbuffer;
static rtpbuffer rtpbuf;

static long TimeToLimit(void);
static void FreeRTPPacket(RTPpacket_t *p);
static void RTPCacheReset(int seq);
static int RTPReceivePacket(RTPpacket_t **pp);
static int RTPReceiveAndCache(RTPpacket_t **pp);
static int RTPGetNext(RTPpacket_t **pp);
static int RTPGetPacketFromNetwork(RTPpacket_t **pp);


#endif

int RTPReadPacket (RTPpacket_t *p, FILE *bits);


/*!
 ************************************************************************
 * \brief
 *    Opens the bit stream file named fn
 * \return
 *    none
 ************************************************************************
 */
void OpenRTPFile (char *fn)
{
  if (NULL == (bits=fopen(fn, "rb")))
  {
    snprintf (errortext, ET_SIZE, "Cannot open RTP file '%s'", input->infile);
    error(errortext,500);
  }
}


/*!
 ************************************************************************
 * \brief
 *    Closes the bit stream file
 ************************************************************************
 */
void CloseRTPFile()
{
  fclose (bits);
}


/*!
 ************************************************************************
 * \brief
 *    Fills nalu->buf and nalu->len with the payload of an RTP packet.  
 *    Other fields in nalu-> remain uninitialized (will be taken care of 
 *    by NALUtoRBSP.
 *
 * \return
 *     4 in case of ok (for compatibility with GetAnnexbNALU)
 *     0 if there is nothing any more to read (EOF)
 *    -1 in case of any error
 *
 ************************************************************************
 */

int GetRTPNALU (NALU_t *nalu)
{
  RTPpacket_t *p;
  int ret;


#if RTP_OVER_NETWORK
  
  if (input->FileFormat == PAR_OF_RTP) {
    if ((p=malloc (sizeof (RTPpacket_t)))== NULL)
      no_mem_exit ("GetRTPNALU-1");
    if ((p->packet=malloc (MAXRTPPACKETSIZE))== NULL)
      no_mem_exit ("GetRTPNALU-2");
    if ((p->payload=malloc (MAXRTPPACKETSIZE))== NULL)
      no_mem_exit ("GetRTPNALU-3");
    
    ret = RTPReadPacket (p, bits);
  }
  else {
    ret = RTPGetPacketFromNetwork(&p);
  }
  
#else
  
  if ((p=malloc (sizeof (RTPpacket_t)))== NULL)
      no_mem_exit ("GetRTPNALU-1");
    if ((p->packet=malloc (MAXRTPPACKETSIZE))== NULL)
      no_mem_exit ("GetRTPNALU-2");
    if ((p->payload=malloc (MAXRTPPACKETSIZE))== NULL)
      no_mem_exit ("GetRTPNALU-3");
    
  ret = RTPReadPacket (p, bits);
  
#endif
  
  nalu->forbidden_bit = 1;
  nalu->len = 0;

  if (ret < 0)
    return -1;
  if (ret == 0)
    return 0;

  assert (p->paylen < nalu->max_size);

  nalu->len = p->paylen;
  memcpy (nalu->buf, p->payload, p->paylen);
  nalu->forbidden_bit = (nalu->buf[0]>>7) & 1;
  nalu->nal_reference_idc = (nalu->buf[0]>>5) & 3;
  nalu->nal_unit_type = (nalu->buf[0]) & 0x1f;

  free (p->payload);
  free (p->packet);
  free (p);
  //  printf ("Got an RTP NALU, len %d, first byte %x\n", nalu->len, nalu->buf[0]);
  return nalu->len;
}



/*!
 *****************************************************************************
 *
 * \brief 
 *    DecomposeRTPpacket interprets the RTP packet and writes the various
 *    structure members of the RTPpacket_t structure
 *
 * \return
 *    0 in case of success
 *    negative error code in case of failure
 *
 * \param p
 *    Caller is responsible to allocate enough memory for the generated payload
 *    in parameter->payload. Typically a malloc of paclen-12 bytes is sufficient
 *
 * \par Side effects
 *    none
 *
 * \date
 *    30 Spetember 2001
 *
 * \author
 *    Stephan Wenger   stewe@cs.tu-berlin.de
 *****************************************************************************/

int DecomposeRTPpacket (RTPpacket_t *p)

{
  // consistency check 
  assert (p->packlen < 65536 - 28);  // IP, UDP headers
  assert (p->packlen >= 12);         // at least a complete RTP header
  assert (p->payload != NULL);
  assert (p->packet != NULL);

  // Extract header information

  p->v  = (p->packet[0] >> 6) & 0x03;
  p->p  = (p->packet[0] >> 5) & 0x01;
  p->x  = (p->packet[0] >> 4) & 0x01;
  p->cc = (p->packet[0] >> 0) & 0x0F;

  p->m  = (p->packet[1] >> 7) & 0x01;
  p->pt = (p->packet[1] >> 0) & 0x7F;

#if RTP_BIG_ENDIAN_SUPPORT
  memcpy (&p->seq, &p->packet[2], 2);
  memcpy (&p->timestamp, &p->packet[4], 4);
  memcpy (&p->ssrc, &p->packet[8], 4);
 
  if (!isBigEndian()) {
    p->seq = ntohs((unsigned short)p->seq);
    p->timestamp = ntohl(p->timestamp);
    p->ssrc = ntohl(p->ssrc);
  }
  
#else

  memcpy (&p->seq, &p->packet[2], 2);
  p->seq = ntohs((unsigned short)p->seq);
  memcpy (&p->timestamp, &p->packet[4], 4);// change to shifts for unified byte sex
  p->timestamp = ntohl(p->timestamp);
  memcpy (&p->ssrc, &p->packet[8], 4);// change to shifts for unified byte sex
  p->ssrc = ntohl(p->ssrc);

#endif
  

  // header consistency checks
  if (     (p->v != 2)
      || (p->p != 0)
      || (p->x != 0)
      || (p->cc != 0) )
  {
    printf ("DecomposeRTPpacket, RTP header consistency problem, header follows\n");
    DumpRTPHeader (p);
    return -1;
  }
  p->paylen = p->packlen-12;
  memcpy (p->payload, &p->packet[12], p->paylen);
  return 0;
}

/*!
 *****************************************************************************
 *
 * \brief 
 *    DumpRTPHeader is a debug tool that dumps a human-readable interpretation
 *    of the RTP header
 *
 * \return
 *    n.a.
 * \param p
 *    the RTP packet to be dumped, after DecompositeRTPpacket()
 *
 * \par Side effects
 *    Debug output to stdout
 *
 * \date
 *    30 Spetember 2001
 *
 * \author
 *    Stephan Wenger   stewe@cs.tu-berlin.de
 *****************************************************************************/

void DumpRTPHeader (RTPpacket_t *p)

{
  int i;
  for (i=0; i< 30; i++)
    printf ("%02x ", p->packet[i]);
  printf ("Version (V): %d\n", p->v);
  printf ("Padding (P): %d\n", p->p);
  printf ("Extension (X): %d\n", p->x);
  printf ("CSRC count (CC): %d\n", p->cc);
  printf ("Marker bit (M): %d\n", p->m);
  printf ("Payload Type (PT): %d\n", p->pt);
  printf ("Sequence Number: %d\n", p->seq);
  printf ("Timestamp: %d\n", p->timestamp);
  printf ("SSRC: %d\n", p->ssrc);
}


/*!
 *****************************************************************************
 *
 * \brief 
 *    RTPReadPacket reads one packet from file
 *
 * \return
 *    0: EOF
 *    negative: error
 *    positive: size of RTP packet in bytes
 *
 * \param p
 *    packet data structure, with memory for p->packet allocated
 *
 * \param bits
 *    target file
 *
 * \par Side effects:
 *   - File pointer in bits moved
 *   - p->xxx filled by reading and Decomposepacket()
 *
 * \date
 *    04 November, 2001
 *
 * \author
 *    Stephan Wenger, stewe@cs.tu-berlin.de
 *****************************************************************************/

int RTPReadPacket (RTPpacket_t *p, FILE *bits)
{
  int Filepos, intime;

  assert (p != NULL);
  assert (p->packet != NULL);
  assert (p->payload != NULL);

  Filepos = ftell (bits);
  if (4 != fread (&p->packlen,1, 4, bits))
  {
    return 0;
  }

  if (4 != fread (&intime, 1, 4, bits))
  {
    fseek (bits, Filepos, SEEK_SET);
    printf ("RTPReadPacket: File corruption, could not read Timestamp, exit\n");
    exit (-1);
  }

#if RTP_BIG_ENDIAN_SUPPORT
  if (!isBigEndian()){
    // Change numbers to little endian format
    p->packlen = ByteSwapLong(p->packlen);
    intime = ByteSwapLong(intime);
  }
#endif
#if 1 // tag: RTP_BIG_ENDIAN_SUPPORT
  p->packlen = ntohl(p->packlen);
#endif

  assert (p->packlen < MAXRTPPACKETSIZE);

  if (p->packlen != fread (p->packet, 1, p->packlen, bits))
  {
    printf ("RTPReadPacket: File corruption, could not read %d bytes\n", p->packlen);
    exit (-1);    // EOF inidication
  }

  if (DecomposeRTPpacket (p) < 0)
  {
    // this should never happen, hence exit() is ok.  We probably do not want to attempt
    // to decode a packet that obviously wasn't generated by RTP
    printf ("Errors reported by DecomposePacket(), exit\n");
    exit (-700);
  }
  assert (p->pt == H26LPAYLOADTYPE);
  assert (p->ssrc == 0x12345678);
  return p->packlen;
}


#if RTP_OVER_NETWORK

/*
 * Returns the difference in milliseconds between the current time and the time to the next
 * timestamp in cache expires. Returns a negative value if the limit has expired.
 */
static long TimeToLimit()
{
  struct timeb now;
  unsigned long timediff, tsdiff;
  long toLimit;
    
  ftime(&now);

  timediff = 1000 * (now.time - timeStart.time) + (now.millitm - timeStart.millitm);
  tsdiff = rtpbuf.nextTimestamp - firstTimestamp;

  // Some not so pretty lines so we're not exceeding variable limits
  // Time limit not exceeded
  if (timediff < tsdiff + RTP_MAX_DELAY)
    toLimit = MIN(-timediff + tsdiff + RTP_MAX_DELAY, LONG_MAX);
  // Time limit exceeded
  else
    toLimit = (signed long) tsdiff + RTP_MAX_DELAY - timediff;

  return toLimit;
}

/*
 * Free resources occupied by the RTP packet
 */
static void FreeRTPPacket(RTPpacket_t *p)
{
  if (p) {
    if (p->payload) free (p->payload);
    if (p->packet) free (p->packet);
    free (p);
  }
}

/*
 * Read data from socket and allocate and create the RTP packet structure. On
 * successfull return PP points to the pointer of this packet. Returns the
 * packet length when successfull, 0 on timeout and a negative value on error.
 */
static int RTPReceivePacket(RTPpacket_t **pp)
{
  struct timeval timeout;
  byte buf[MAXRTPPACKETSIZE];
  int nfd;
  socklen_t socklen;
  RTPpacket_t *p;

  assert(pp);
  
  // Allocate RTP packet
  if ((p=malloc (sizeof (RTPpacket_t)))== NULL)
    no_mem_exit ("RTPReceivePacket-1");
  if ((p->packet=malloc (MAXRTPPACKETSIZE))== NULL)
    no_mem_exit ("RTPReceivePacket-2");
  if ((p->payload=malloc (MAXRTPPACKETSIZE))== NULL)
    no_mem_exit ("RTPReceivePacket-3");

  timeout.tv_sec = LISTEN_TIMEOUT;
  timeout.tv_usec = 0;

  memset(&con.encAddr, 0, sizeof(con.encAddr));
  socklen = sizeof(con.encAddr);

  // Check if socket is ready for reading with a timeout
  nfd = select(con.sockfd+1, &con.readfds, NULL, NULL, &timeout);

  //Receive data and create the packet structure
  if (nfd > 0) {
    p->packlen = recvfrom(con.sockfd, buf, sizeof(buf), 0,
                          (struct sockaddr *) &con.encAddr, &socklen);
    
#if 0
    p->packlen -= 8;
    memcpy(p->packet, &buf[8], p->packlen);
#endif
    memcpy(p->packet, buf, p->packlen);
    
    if (DecomposeRTPpacket(p) < 0) {
      fprintf (stderr, "Errors reported by DecomposePacket()\n");
      FreeRTPPacket(p);
      return -1;
    }

    /* Write received data (undecoded) to file */
    int intime = -1;
    int packlen_n = htonl(p->packlen);
    write(p_received, &packlen_n, 4);
    write(p_received, &intime, 4);
    write(p_received, p->packet, p->packlen);
  }
  // We got a timeout. The end of the received video reached (most likely).
  else if (nfd == 0) {
    rtcp_alive = 0;
    FreeRTPPacket(p);
    close(p_received);
    return 0;    
  }
  else {
    fprintf(stderr, "An error occured while listening for RTP packets\n");
    FreeRTPPacket(p);
    close(p_received);
    return -2;
  }

  *pp = p;

  assert (p->pt == H26LPAYLOADTYPE);
  assert (p->ssrc == 0x12345678);
  return p->packlen;
}


/*
 * Reset cache. SEQ is the seq no of last received packet.
 */
static void RTPCacheReset(int seq)
{
  int i;

  rtpbuf.first = 0;
  rtpbuf.seq[0] = seq;

  
  for (i = 0; i < MAX_RTPPACKETS_IN; i++) {
    rtpbuf.len[i] = 0;
    if (rtpbuf.pack[i]) {
      FreeRTPPacket(rtpbuf.pack[i]);
    }
  }
}

/*
 * Receive and write an RTP packet in cache in correct sequence order.
 *
 * If packet received is "next in sequence" (or cache is reset), the length of
 * the packet is returned. In this situation the actual packet is not stored in
 * cache, but information that it has been received is. Instead the pointer PP
 * will point to a pointer of the packet, which should be used to reach the
 * packet.
 *
 * If another packet than the next in sequence is returned (eg. out of seq,
 * duplicate etc.), 0 is returned and the value of PP is uncertain.
 *
 * If no packet was received (and cached) due to "normal" circumstances -1 is
 * returned. If other errors occured, another negative value is returned.
 */
static int RTPReceiveAndCache(RTPpacket_t **pp)
{
  // todo: (low pri) handle seq cycle (occurs for every 2^16 packet)
  
  int len;
  int newseq;
  unsigned short seq;
  static short is_first = 1;
  int i;

  assert(pp);
    
  len = RTPReceivePacket(pp);
  
  // timeout
  if (len == 0) {
    return -1;
  }
  // error
  else if (len < 0) {
    fprintf(stderr, "Receive packet from network failed in RTPReceiveAndCache(), exit\n");
    exit (-2);
  }

  if (is_first) {
    // set reference time
    ftime(&timeStart);
    firstTimestamp = (*pp)->timestamp;
    rtpbuf.nextTimestamp = RTP_CACHE_EMPTY_TIMESTAMP;
    // start RTCP feedback
    rtcp_init((*pp)->seq);
  }

  /* All packets should be reported for RTCP, including late and duplicate
   * packages (RFC3550 sec 6.4.1 (cumulative number of packets lost)) */
  rtcp_rtp_packet_received(*pp);

  seq = (*pp)->seq;
  newseq = seq - rtpbuf.seq[rtpbuf.first];

  // received seq no is in sequential order (or first packet)
  if ((newseq == 0) || is_first) {
    is_first = 0;
    RTP_DEBUG(("In sequence (seq[%d]=%d seq=%d, newseq=%d)\n",
             rtpbuf.first, rtpbuf.seq[rtpbuf.first], seq, newseq));
    rtpbuf.first = (rtpbuf.first + 1) % MAX_RTPPACKETS_IN;
    rtpbuf.seq[rtpbuf.first] = seq + 1;

    return len;
  }

  // Received seq no is too large for buffer. Indicates many lost packets.
  // Reset buffer and return this packet
  if (newseq >= MAX_RTPPACKETS_IN)
  {
    RTP_DEBUG(("Overrun (seq[%d]=%d seq=%d, newseq=%d)\n",
             rtpbuf.first, rtpbuf.seq[rtpbuf.first], seq, newseq));

    uint16_t j = rtpbuf.seq[rtpbuf.first];
    while (j < seq) {
      rtcp_rtp_packet_lossed(j);
    }
    
    RTPCacheReset(seq+1);

    return len;
  }

  // Received seq no was smaller than expected
  if (newseq < 0) {

    // Check for duplicate. If so, ignore it (but report to RTCP for
    // statistical purposes).
    for (i = 0; i < MAX_RTPPACKETS_IN; i++) {
      if (rtpbuf.seq[i] == seq) {
        RTP_DEBUG(("Duplicate packet (seq[%d]=%d seq=%d, newseq=%d found at %d)\n",
                 rtpbuf.first, rtpbuf.seq[rtpbuf.first], seq, newseq, i));
        FreeRTPPacket(*pp);
        return 0;
      }
    }

    // Packet is too old.
    RTP_DEBUG (("Too Old packet (seq[%d]=%d seq=%d, newseq=%d)\n",
             rtpbuf.first, rtpbuf.seq[rtpbuf.first], seq, newseq));
    FreeRTPPacket(*pp);
    return 0;
  }

  // Seq no is a little bigger than expected.
  // Out of seq. ( 0 < newseq < MAX_RTPPACKETS_IN )
  // Cache packet in correct postition.
//  RTP_DEBUG (("Out of Seq (seq[%d]=%d seq=%d, newseq=%d)\n",
//            rtpbuf.first, rtpbuf.seq[rtpbuf.first], seq, newseq));
  RTP_DEBUG (("Out of Seq (seq[%d]=%d seq=%d, newseq=%d, cached[%d])\n",
            rtpbuf.first, rtpbuf.seq[rtpbuf.first], (*pp)->seq, newseq,
            (newseq + rtpbuf.first) % MAX_RTPPACKETS_IN));

  newseq = (newseq + rtpbuf.first) % MAX_RTPPACKETS_IN;
  rtpbuf.pack[newseq] = *pp;
  rtpbuf.len[newseq] = len;
  rtpbuf.seq[newseq] = seq;
  if ((*pp)->timestamp < rtpbuf.nextTimestamp)
    rtpbuf.nextTimestamp = (*pp)->timestamp;
  

  if (seq - rtpbuf.seq[rtpbuf.first] == MAX_RTPPACKETS_IN - 1) {
    RTP_DEBUG(("Cache is full\n"));
    return -2;      
  }
  
  return 0;
}

/*
 * Get the next packet from the RTP cache in sequence. Receive a new packet from
 * the network if necessary. Returns the length of the packet. Should be larger
 * than 0. Returns -1 on timeout.
 */
static int RTPGetNext(RTPpacket_t **pp)
{
  int i, j, len;
  unsigned short nextseq;
  static short is_first = 1;

  assert(pp);

  /* Iterate through this loop and fill cache until we got the next packet or
   * until we decide the packet is lost. */
  for (i=0; i < MAX_RTPPACKETS_IN; i++) {

    // Got next packet in cache
    if (rtpbuf.len[rtpbuf.first] != 0)
      break;

    // Check if time limit is expired on packets in cache.
    long tmp;
    if (!is_first && (tmp = TimeToLimit()) < 0) {
      RTP_DEBUG(("Time limit for packet in cache expired with %li ms\n", tmp));
      break;
    }
    is_first = 0;

    // Get a packet and put in cache
    len = RTPReceiveAndCache(pp);

    // return if packet is the next in sequence
    if (len > 0){
      RTP_DEBUG(("Got RTP packet from network, seq=%hu\n", (*pp)->seq));
      return len;
    }
    
    // Timeout
    else if (len == -1) {
      RTP_DEBUG(("Timeout indicated at RTP %hu [%d]\n",
                rtpbuf.seq[rtpbuf.first], rtpbuf.first));
      return -1;
    }

    // Cache full
    else if (len == -2)
      break;
    
    // Should not happen
    else if (len < -2) {
      RTP_DEBUG(("Unexpected return value '%d' from RTPReceiveAndCache() (i=%d [%hu])\n",
                len, i, rtpbuf.first));
      break;
    }
    // Only if len == 0 loop continues!
  }

  /* Got next packet in cache or the packet expected to be next in sequence did
   * not arrive */
  
  // Find position in cache to get the next packet from
  i = rtpbuf.first;
  j = 0;
  while (rtpbuf.len[i] == 0) {
    int seq = rtpbuf.seq[rtpbuf.first] + j++;
    printf("Lost packet %hu\n", seq);
    rtcp_rtp_packet_lossed(seq);
    i = (i + 1) % MAX_RTPPACKETS_IN;
    if (i == rtpbuf.first)
      break;
  }
  rtpbuf.first = i;

  // Get packet from cache
  *pp = rtpbuf.pack[rtpbuf.first];
  len = rtpbuf.len[rtpbuf.first];
  RTP_DEBUG(("Got packet from cache[%d], seq=%hu \n",
           rtpbuf.first, (*pp)->seq));
  

  // Reset slot and go next in cache
  rtpbuf.len[rtpbuf.first] = 0;
  rtpbuf.pack[rtpbuf.first] = NULL;
  nextseq = rtpbuf.seq[rtpbuf.first];
  rtpbuf.first = ( 1 + rtpbuf.first ) % MAX_RTPPACKETS_IN;
  rtpbuf.seq[rtpbuf.first] = nextseq + 1;
  
  i = rtpbuf.first;
  while (rtpbuf.len[i] == 0) {
    i = (i + 1) % MAX_RTPPACKETS_IN;
    if (i == rtpbuf.first)
      break;
  }

  if (i != rtpbuf.first)
    rtpbuf.nextTimestamp = rtpbuf.pack[i]->timestamp;
  else
    // cache empty, set nextTimestamp to default value
    rtpbuf.nextTimestamp = RTP_CACHE_EMPTY_TIMESTAMP;
    

  return len;  
}


/*
 * Read the next RTP packet using cache. Returns length of packet, or 0 if
 * timeout occured.
 */
static int RTPGetPacketFromNetwork(RTPpacket_t **pp)
{
  int len;
  
  assert(pp);

  // Loop to skip empty packets
  while ((len = RTPGetNext(pp)) == 0)
    RTP_DEBUG(("Got empty packet from RTP cache.\n"));

  if (len == -1)
    return 0;

  return len;
}


void OpenRTPConnection(unsigned short port)
{
  /* Create a socket and bind it */
  
  memset(&con.decAddr, 0, sizeof(con.decAddr));
  con.decAddr.sin_family = AF_INET;
  con.decAddr.sin_addr.s_addr = htonl(INADDR_ANY);
  con.decAddr.sin_port = htons(port);

  if ((con.sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) 
  {
    perror("Server: Failed creating socket");
    exit(-1);
  }

  if ((bind(con.sockfd, (struct sockaddr *)&con.decAddr, sizeof(con.decAddr)) < 0)) 
  {
    perror("Server: Failed bind");
    exit(-1);
  }

  /* Prepare for using 'select' when listening */
  FD_ZERO(&con.readfds);
  FD_SET(con.sockfd, &con.readfds);

  RTPCacheReset(0);
  
  /* Also open a file to write the received raw data . We use the filename
   * defined as infile because this is not used anyway in this mode. */
  if ((p_received=open(input->infile, OPENFLAGS_WRITE, OPEN_PERMISSIONS))==-1)
  {                                     
    snprintf(errortext, ET_SIZE, "Error open file %s ",input->infile);
    error(errortext,500);
  }
}

#endif
