#include "defines.h"
#if FBIR

#ifndef _RTCP_H_
#define _RTCP_H_

#define RTCP_DEBUG_ENABLE 1

#include <stdint.h>
#include <sys/socket.h>
#include <arpa/inet.h>

typedef enum {
  /* RTCP packet types according to
   * http://www.iana.org/assignments/rtp-parameters */
  RTCP_FIR   = 192,
  RTCP_NACK  = 193,
  RTCP_SR    = 200,
  RTCP_RR    = 201,
  RTCP_SDES  = 202,
  RTCP_BYE   = 203,
  RTCP_APP   = 204,
  RTCP_RTPFB = 205,
  RTCP_PSFB  = 206,
  RTCP_XR    = 207
} rtcp_packet_types;

//#define RTCP_DEFAULT_PORT 62003
#define RTCP_MAX_PACKET_SIZE (65536 - 28) /* Max size IP packet - IP header - UDP header */
#define RTCP_MAX_SDES_ITEM_LENGTH 255


typedef struct {
  uint32_t ssrc_n;                 /* ssrc, 32 bits */
  uint8_t fl;                    /* fraction lost, 8 bits */
  uint32_t cnlp;                   /* cumulative number of lost packets,
                                         * 24 bits */
  uint32_t ehsn;                   /* extended highest seq no, 32 bits */
  uint32_t jitter;                 /* interarrival jitter, 32 bits */
  uint32_t lsr;                    /* last SR, 32 bits */
  uint32_t dlsr;                        /* dealy since last SR, 32 bits */
} rtcp_rr_report_block_t;


typedef struct {
  uint8_t v;                     /* version, 2 bits */
  uint8_t p;                     /* padding, 1 bit */
  uint8_t rc;                    /* rr count, 5 bits */
  uint8_t pt;                    /* packet type, 2 bits */
  uint16_t length;                  /* length of packet, 16 bits */
  uint32_t ssrc;                   /* ssrc, 32 bits */
  rtcp_rr_report_block_t *rb;           /* report block 1. According to the
                                         * standard this should be an array,
                                         * but we only need one. */
} rtcp_rr_t;

/* An extremely simplified struct for a SDES message which only can contain
 * one SDES item, namely the mandatory CNAME item. */
typedef struct {
  uint8_t v;
  uint8_t p;
  uint8_t sc;
  uint8_t pt;
  uint16_t length;
  uint32_t chunk_csrc;
  uint8_t chunk_item_type;
  uint8_t chunk_item_length;
  char *chunk_item_cname;                
} rtcp_sdes_t;

typedef struct {
  uint16_t pid;                     /* RTP seq of lost packet, 16 bits */
  uint16_t blp;                     /* Bitmask of following lost packets,
                                         * 16 bits */
} rtcp_fci_nack_t;

typedef struct {
  uint8_t v;                            /* version, 2 bits */
  uint8_t p;                            /* padding, 1 bit */
  uint8_t fmt;                          /* feedback message type, 5 bits */
  uint8_t pt;                           /* payload type, 8 bits */
  uint16_t length;                      /* length of packet, 16 bits */
  uint32_t ssrc_sender;                 /* ssrc of the sender of this packet,
                                         * 32 bits*/
  uint32_t ssrc_source;                 /* ssrc of the related media source,
                                         * 32 bits */
  rtcp_fci_nack_t *fci;               /* list feedback control information */
  int n_fci;                            /* coint elements in fci list */
} rtcp_fb_t;

typedef struct {
  rtcp_rr_t *rr;
  rtcp_sdes_t *sdes;
  rtcp_fb_t *fb;                        /* Since we only have one type of FB,
                                           only one fb entry is required.*/
  short has_rr;
  short has_sdes;
  short has_fb;

  byte *packet;
  int packlen;
} rtcp_packet_t;

typedef struct {
  unsigned short port;                  /* Port number receiving data on */
  struct sockaddr_in decaddr;           /* Decoders socket information */
  struct sockaddr_in encaddr;           /* Encoders socket information */
  int sockfd;                           /* Socket id */
  fd_set readfds;
} rtcp_connection_t;


void OpenRTCPConnection(unsigned short port);
void CloseRTCPConnection();

#endif
#endif
