
/*!
 ***************************************************************************
 *
 * \file rtp.h
 *
 * \brief
 *    Definition of structures and functions to handle RTP headers.  For a
 *    description of RTP see RFC1889 on http://www.ietf.org
 *
 * \date
 *    30 September 2001
 *
 * \author
 *    Stephan Wenger   stewe@cs.tu-berlin.de
 **************************************************************************/

#ifndef _RTP_H_
#define _RTP_H_

#include "nalu.h"
#include <sys/socket.h>
#include <arpa/inet.h>

#if GILBERT_IN_ENCODER
#include "gilbert.h"
#endif


#define MAXRTPPAYLOADLEN  (65536 - 40)    //!< Maximum payload size of an RTP packet
#define MAXRTPPACKETSIZE  (65536 - 28)    //!< Maximum size of an RTP packet incl. header
#define H26LPAYLOADTYPE 105               //!< RTP paylaod type fixed here for simplicity
#define H26LSSRC 0x12345678               //!< SSRC, chosen to simplify debugging
#define RTP_TR_TIMESTAMP_MULT 1000        //!< should be something like 27 Mhz / 29.97 Hz

typedef struct 
{
  unsigned int v;          //!< Version, 2 bits, MUST be 0x2
  unsigned int p;          //!< Padding bit, Padding MUST NOT be used
  unsigned int x;          //!< Extension, MUST be zero */
  unsigned int cc;         /*!< CSRC count, normally 0 in the absence 
                                of RTP mixers */
  unsigned int m;          //!< Marker bit
  unsigned int pt;         //!< 7 bits, Payload Type, dynamically established
  unsigned int seq;        /*!< RTP sequence number, incremented by one for 
                                each sent packet */
#if RTP_OVER_NETWORK
  unsigned long timestamp;              /* be sure it is at least 32 bits */
#else
  unsigned int timestamp;  //!< timestamp, 27 MHz for H.264
#endif
  unsigned int ssrc;       //!< Synchronization Source, chosen randomly
  byte *       payload;    //!< the payload including payload headers
  unsigned int paylen;     //!< length of payload in bytes
  byte *       packet;     //!< complete packet including header and payload
  unsigned int packlen;    //!< length of packet, typically paylen+12
} RTPpacket_t;

typedef struct {
  char recvHost[256];
  unsigned short recvPort;
  int sockfd;
  struct sockaddr_in recvAddr;
} RTPConnection;

#if 0
int  ComposeRTPPacket (RTPpacket_t *p);
int  DecomposeRTPpacket (RTPpacket_t *p);
int  WriteRTPPacket (RTPpacket_t *p, FILE *f);
#if GILBERT_IN_ENCODER
int  WriteRTPPacket (RTPpacket_t *p, FILE *f);
#endif
void DumpRTPHeader (RTPpacket_t *p);
void RTPUpdateTimestamp (int tr);
int  RTPWriteBits (int Marker, int PacketType, void * bitstream, 
                   int BitStreamLenInByte, FILE *out);

Boolean isAggregationPacket();
int aggregationRTPWriteBits (int Marker, int PacketType, int subPacketType, void * bitstream, int BitStreamLenInByte, FILE *out);

void begin_sub_sequence_rtp();
void end_sub_sequence_rtp();
#endif

void RTPUpdateTimestamp (int tr);
void OpenRTPFile (char *Filename);
void CloseRTPFile ();
int WriteRTPNALU (NALU_t *n);
#if GILBERT_IN_ENCODER
void OpenRTPFileLossy (char *Filename);
void CloseRTPFileLossy ();
int WriteRTPNALULossy (NALU_t *n);
#endif
#if RTP_OVER_NETWORK
void OpenRTPConnection(const char *host, unsigned short port);
void CloseRTPConnection();
int SendRTPNALU(NALU_t *n);
#endif


#endif

