#include "defines.h"

#if FBIR

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <string.h>

#include "fbir.h"
#include "global.h"
#include "image.h"
#include "memalloc.h"
#include "minmax.h"


static short ****mv_mem;                /* mv history [memidx][subblock_y][subblock_x][component] */
static int *corrupted;
static int *intraMB;
static int checked_mbs;            /* 0 or positive when mbs needs to be
                                         * checked */
static FILE *fbir_log;

#define NO_CORRUPTION INT_MAX
static int oldest_corruption = NO_CORRUPTION;

#define PIX_TO_BLOCK(x) (x >> 2)
#define PIX_TO_MB(x, y) ((y/16)*(img->width/16) + (x/16))

#if ET_DEBUG_ENABLE
#define ET_DEBUG(s) printf s
#else
#define ET_DEBUG(s)
#endif


/*
 * Init variables and open files for FBIR
 */
void fbir_init()
{
  int i;
  int number_of_mbs = img->height * img->width / 256;

  /* init memory */
  get_mem4Dshort(&mv_mem, MV_MEMORY_SIZE, img->height / BLOCK_SIZE, img->width / BLOCK_SIZE, 2);

  if ((corrupted = (int *) malloc(sizeof(int) * number_of_mbs)) == NULL)
    no_mem_exit("fbir_init(): corrupted");
  
  if ((intraMB = (int *) malloc(sizeof(int) * number_of_mbs)) == NULL)
    no_mem_exit("fbir_init(): intraMB");

  for (i = 0; i < number_of_mbs; i++) {
    corrupted[i] = -1;
    intraMB[i] = -1;
  }
  
  checked_mbs = -1;

  /* init i/o */
  if(strlen(input->FBIntraRefreshLogFile) > 0 && (fbir_log = fopen(input->FBIntraRefreshLogFile, "wt")) == NULL) {
    fprintf(stderr, "Error open file '%s'\n", input->FBIntraRefreshLogFile);
    exit(-1);
  }

  /* init stats */
  fbir_stats.lost = 0;
  fbir_stats.corr = 0;
  fbir_stats.corr_total = 0;
  fbir_stats.corr_under_limit = 0;
  fbir_stats.refreshed = 0;
  fbir_stats.static_ = 0;
  fbir_stats.avg_refresh_time = 0;
}

/*
 * Free all variables used by FBIR
 */
void fbir_free()
{
  int buflen = 2048;
  char buf[buflen];
  char *bp = buf;
  char *p = bp;
  
  /* Write stats to end of file */
  p += snprintf(p, buflen - (p - bp), "\n------------------------------------\n");
  p += snprintf(p, buflen - (p - bp), "MBs lost: %u\n", fbir_stats.lost);
  p += snprintf(p, buflen - (p - bp), "Corrupted MBs refreshed: %u (avg.time: %.2f)\n",
                fbir_stats.refreshed, fbir_stats.avg_refresh_time);
  p += snprintf(p, buflen - (p - bp), "Corrupted static MBs: %u\n", fbir_stats.static_);
  p += snprintf(p, buflen - (p - bp), "Corrupted MBs with 'CR < limit': %u\n",
                fbir_stats.corr_under_limit);

  fbir_log_write(buf);
  fclose(fbir_log);

  free_mem4Dshort(mv_mem, MV_MEMORY_SIZE, img->height / BLOCK_SIZE);
  free(corrupted);
  free(intraMB);
}


/*
 * Write string S to FBIR log file
 */
void fbir_log_write(char *s)
{
  fprintf(fbir_log, s);
}

/*
 * Track error for pixel in frame number PIC and pixel (PIX_X, PIX_Y)
 *
 * Return 0 if no error, or a postive value representing the number of frames
 * since this pixel was corrupted
 */
static int track_error_pixel (int pic, int pix_x, int pix_y)
{
  assert(oldest_corruption != NO_CORRUPTION);

  int z_last;
  int z = pic % MV_MEMORY_SIZE;
  int block_y, block_x;
  int ref_pic = pic - 1;

  if (pic - oldest_corruption < MV_MEMORY_SIZE) 
    z_last = oldest_corruption % MV_MEMORY_SIZE;
  else
    z_last = (z == MV_MEMORY_SIZE-1) ? 0 : z + 1;
  
  while (z != z_last) {
    /* Trace where the pixel was predicted from */
    block_x = PIX_TO_BLOCK(pix_x);
    block_y = PIX_TO_BLOCK(pix_y);
    pix_x = pix_x + (int)(mv_mem[z][block_y][block_x][0] / 4.0 + 0.5);
    pix_y = pix_y + (int)(mv_mem[z][block_y][block_x][1] / 4.0 + 0.5);

 
    /* If pixel references a place outside the image it is correct ? */
    if (pix_x < 0 || pix_y < 0 ||
      pix_x >= img->width || pix_y >= img->height)
      return 0;

    /* Check if referenced area is corrupted */
    int mb = PIX_TO_MB(pix_x, pix_y);
    if (corrupted[mb] == ref_pic)
      return pic - ref_pic;
      
    if (intraMB[mb] == ref_pic)
      return 0;
    
    z = (z == 0) ? MV_MEMORY_SIZE-1 : z-1;
    ref_pic--;  
  }
  /* If we don't have stored MVs back to PIC and not concluded its status
   * yet, set it as corrupted because we don't know. (Better to refresh too
   * much than not recover the error)*/
  if (pic - oldest_corruption > MV_MEMORY_SIZE-1)
    return MV_MEMORY_SIZE;

  return 0;
}


/*
 * Returns corruption ratio (0-256) of MB in PIC
 */
static int track_error (int pic, int mb, short pmv[2])
{
  assert(oldest_corruption != NO_CORRUPTION);

  ET_DEBUG(("Tracking error of MB %d in PIC %d\n", mb, pic));
  
  int z, y, x, i, j;
  int tl_x, tr_x, bl_x, br_x;
  int tl_y, tr_y, bl_y, br_y;
  int block_y = (mb / (img->width / 16)) << 2;
  int block_x = (mb % (img->width / 16)) << 2;
  //short pmv[2];
  int cr = 0;
  int corrtime = 0, maxcorrtime = 0;

  z = pic % MV_MEMORY_SIZE;
  tl_x = bl_x = block_x << 2;
  tr_x = br_x = tl_x + 15;
  tl_y = tr_y = block_y << 2;
  bl_y = br_y = tl_y + 15;
  
  for (i = 0; i < 4; i++) {
    y = block_y + i;
    for (j = 0; j < 4; j++) {
      x = block_x + i;
      mv_mem[z][y][x][0] = pmv[0];
      mv_mem[z][y][x][1] = pmv[1];
    }
  }
  
  /* Check if corners of MB is in corrupted region */
  if ((corrtime = track_error_pixel(pic, tl_x, tl_y))) {
    cr ++;
    maxcorrtime = max(corrtime, maxcorrtime);
  }
  if ((corrtime = track_error_pixel(pic, tr_x, tr_y))) {
    cr ++;
    maxcorrtime = max(corrtime, maxcorrtime);
  }
  if ((corrtime = track_error_pixel(pic, bl_x, bl_y))) {
    cr ++;
    maxcorrtime = max(corrtime, maxcorrtime);
  }
  if ((corrtime = track_error_pixel(pic, br_x, br_y))) {
    cr ++;
    maxcorrtime = max(corrtime, maxcorrtime);
  }

  /* Only calculate full CR if one or more of the corners is corrupted */
  if (cr == 0)
    return 0;

  /* Calculate corruption of rest of pixels in MB */
  for (y = tl_y; y <= bl_y; y++) {
    for (x = tl_x; x <= tr_x; x++) {
      if (
        (x == tl_x && y == tl_y) ||
        (x == tr_x && y == tr_y) ||
        (x == bl_x && y == bl_y) ||
        (x == br_x && y == br_y))
        continue;
      if ((corrtime = track_error_pixel(pic, x, y))) {
        cr ++;
        maxcorrtime = max(corrtime, maxcorrtime);
      }
    }
  }

  fbir_stats.last_fb_time = maxcorrtime;
  return cr;
}


/*
 * Returns the CR value (contamination ratio) between 0 and 256 for the mb
 * number MB in frame PIC. 
 */
int mb_corruption (int pic, int mb, int intra, int last_non_static)
{
  int i;
  int cr;
  short pmv[2];
  int number_of_mbs = img->height * img->width / 256;

  /* No errors that hasn't already been accounted for */
  if (checked_mbs < 0)
    return CR_NO_CORRUPTED_MBS;
  else if(checked_mbs == 0 && mb != 0)
    return CR_CORRUPTION_WAITING;

  checked_mbs++;
  if (intra) {  
    /* No need to check */
    cr = 0;
  }
  else if (checked_mbs > number_of_mbs) {
    /* All MBs checked. Reset */
    for (i = 0; i < number_of_mbs; i++)
      corrupted[i] = -1;
    checked_mbs = -1;
    oldest_corruption = NO_CORRUPTION;
    cr = CR_ALL_MBS_CHECKED;
  }
  else {
    if (corrupted[mb] != -1) {
      /* Check if MB's "skip history" indicates this MB as part of a static
       * area. If so, let CR = 0 as the decoder most likely has recovered the
       * error well. */
      int static_ = (last_non_static <  corrupted[mb]) ? 1 : 0;
    
      /* We don't have calculated MVs for current picture yet. (This function
       * is called before that happens.) Therefore we use predicted MVs
       * instead. */
      SetMotionVectorPredictor(pmv, enc_picture->ref_idx[LIST_0],
                               enc_picture->mv[LIST_0], 0, LIST_0,
                               0, 0, 16, 16);
    
      if (static_ &&
          abs(pmv[0]) <= MAX_MOTION_WHEN_STATIC &&
          abs(pmv[1]) <= MAX_MOTION_WHEN_STATIC) {
        /* Skiphist indicates all MBs since error as skip  */
        ET_DEBUG(("MB %d been static since %d (last_non_static=%d, loss in frame=%d)\n",
                  mb, last_non_static[mb]+1, last_non_static[mb], corrupted[mb]));
        return CR_STATIC_MB;
      }
    }
    cr = track_error(pic, mb, pmv);    
  }

  return cr;
}

/*
 * Store variables for current MB which will be used for error tracking if
 * necessary. INTRA is 0 if the MB is NOT intra coded */
void et_store_info (int intra)
{
  int z, y, x, i, j;
  int block_y = img->block_y;
  int block_x = img->block_x;
  int pic = img->number;
  int mb = img->current_mb_nr;
  short ***mv = enc_picture->mv[LIST_0]; /* Assume no B-slices */
  
  if (intra)
    intraMB[mb] = pic;

  ET_DEBUG(("Store MVs for MB %d in #%d:\n", img->current_mb_nr, img->number));
  ET_DEBUG(("MB info: b8mode = {%d, %d, %d, %d}\n", mb_curr->b8mode[0], mb_curr->b8mode[1], mb_curr->b8mode[2], mb_curr->b8mode[3]));
  z = pic % MV_MEMORY_SIZE;


#if ET_DEBUG_ENABLE
  short pmv[2];
  SetMotionVectorPredictor(pmv,
                           enc_picture->ref_idx[LIST_0],
                           enc_picture->mv[LIST_0],
                           0,
                           LIST_0,
                           0,
                           0,
                           16,
                           16);
#endif
  
  for (i = 0; i < 4; i++) {
    y = block_y + i;
    for (j = 0; j < 4; j++) {
      x = block_x + j;
      mv_mem[z][y][x][0] = mv[y][x][0];
      mv_mem[z][y][x][1] = mv[y][x][1];
//      ET_DEBUG(("\tmv_mem[%d][%d][%d][0] = %d\n", z, y, x, mv_mem[z][y][x][0]));
//      ET_DEBUG(("\tmv_mem[%d][%d][%d][1] = %d\n", z, y, x, mv_mem[z][y][x][1]));
      ET_DEBUG(("\tmv_mem[%d][%d][%d][0] = %d    (pred = %d)\n", z, y, x, mv_mem[z][y][x][0], pmv[0]));
      ET_DEBUG(("\tmv_mem[%d][%d][%d][1] = %d    (pred = %d)\n", z, y, x, mv_mem[z][y][x][1], pmv[1]));
    }
  }
}


/*
 * Register RTP packet with sequence number SEQ as lost. Variables will be
 * set so that other functions know to perform error tracking
 */
void register_lost_packet (int seq)
{
  char buf[512];

  int pic;
  int slice;
  int mb;
  int i;
  int const firstSeq = 0;               /* Init value of
                                         * CurrentRTPSequenceNumber in rtp.c */
  int const mbs_per_slice = input->slice_argument;
  /* Assure that an integer number of slices covers the whole picture */
  int const mbs_per_pic = img->width * img->height / 256;  
  assert((float) mbs_per_pic / mbs_per_slice == mbs_per_pic / mbs_per_slice);
  int const slices_per_pic = mbs_per_pic / mbs_per_slice;
  
  /* We make some assumptions so it's easy to calculate the MBs contained
   * in RTP packet seq. */
  assert(input->slice_mode == FIXED_MB);
  assert(input->ResendPPS == 0);
  assert(input->GenerateMultiplePPS == 0);

  snprintf(buf, sizeof(buf), "Packet %d is reported lost.\n", seq);
  fbir_log_write(buf);
  
  /* There is one slice in each RTP packet and a fixed number of slices in
   * each picture, hence the picture and slice number packet SEQ belongs to can be
   * calculated as follows */
  
  /* todo: (low pri) support for cycling sequence numbers and negative pic */
  /* '-2' because the two first packets contains parameter sets. Ugly!!!  */
  pic = (seq - firstSeq - 2) / slices_per_pic;
  slice = (seq - firstSeq - 2) % slices_per_pic;

  /* Give warning if loss is in parameter set or first frame */
  if (pic < 1)
    fprintf(stderr, "Warning: Loss in parameter set or first frame.\n");
  
  oldest_corruption = (pic < oldest_corruption) ? pic : oldest_corruption;


  for (i = 0; i < mbs_per_slice; i++){
    mb = slice * mbs_per_slice + i;
    corrupted[mb] = pic;
    snprintf(buf, sizeof(buf), "\tMB %d in PIC %d is lost.\n", mb, pic);
    fbir_log_write(buf);
    fbir_stats.lost ++;
  }
  
  /* When checked_mbs is set 0, the error tracking function (mb_corruption())
   * knows that there are MBs that need tracking an error. */
  checked_mbs = 0;
}

#endif
