#include "defines.h"
#if GILBERT_IN_ENCODER

#include <time.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "global.h"
#include "gilbert.h"

static gilbertStat gStat;
static gilbertModel gModel;

/*
 * Init loss statistics
 */
static void initGilbertStat()
{
  gStat.nPackets = 0;
  gStat.nPacketsLost = 0;
  gStat.nBursts = 0;
  gStat.avgBurstLength = 0;
  gStat.currentBurstLength = 0;
  if (strlen(input->GilbertLogFile) > 0 && (gStat.file = fopen(input->GilbertLogFile, "wt")) == NULL){
    snprintf(errortext, ET_SIZE, "Error: Could not open Gilbert log file `%s'", input->GilbertLogFile);
    error (errortext, 500);
  }
}

/*
 * Update loss statistics.
 */
static void updateGilbertStat(int packetnr, int state)
{
  static int prevState = NON_LOSSY;

  gStat.nPackets++;
  if (state == LOSSY && prevState == NON_LOSSY) {
    gStat.nPacketsLost++;
    gStat.nBursts++;
    gStat.currentBurstLength = 1;
    prevState = state;
  }
  else if (state == LOSSY && prevState == LOSSY) {
    gStat.nPacketsLost++;
    gStat.currentBurstLength++;
  }
  else if (state == NON_LOSSY && prevState == LOSSY) {
    gStat.avgBurstLength = ( (gStat.nBursts-1) * gStat.avgBurstLength + gStat.currentBurstLength ) / gStat.nBursts;
    gStat.currentBurstLength = 0;
    prevState = NON_LOSSY;
  }

  if (state == LOSSY)
    fprintf(gStat.file, "Packet %6d was lost\n", packetnr);
}


/*
 * Free statistic variables and write summary to log file
 */
static void freeGilbertStat()
{
  fprintf(gStat.file, "\n----------------------------------\n");
  fprintf(gStat.file, "%d out of %d packets lost\n", gStat.nPacketsLost, gStat.nPackets);
  fprintf(gStat.file, "Average loss rate: %.2f. (Expected was %d)\n",
          (float) gStat.nPacketsLost/gStat.nPackets * 100, input->GilbertLossRate);
  fprintf(gStat.file, "%d bursts with an average length of %.2f. (Expected was %d)\n",
          gStat.nBursts, gStat.avgBurstLength, input->GilbertBurstLength);
  fclose(gStat.file);
}


/*
 * Init Gilbert model variables
 */
void initGilbertModel()
{
  int lr = input->GilbertLossRate;
  int bl =  input->GilbertBurstLength;
  gModel.q = (float) 1/bl;  /* transition prob. from lossy to non-lossy state */
  gModel.p = (float) lr / (bl * (100-lr) );
  gModel.state = NON_LOSSY;

  /* set initial seed */
  srand((unsigned) rand() + time(NULL));

  initGilbertStat();
}

/*
 * Free all variables for the Gilbert model
 */
void freeGilbertModel()
{
  freeGilbertStat();
}


/*
 * Simulates a lossy network using the gilbert model. Must be called after
 * initGilbertModel(). Returns 0 if packet is lost and 1 if not.
 *
 * Seq is the sequence number of the packet which is sent. Used for logging
 * purposes.
 */
int simLossyNetwork(int seq)
{
  float r = (float) rand()/RAND_MAX;
    
  switch(gModel.state)
  {
    case NON_LOSSY:                     /* in non-lossy state */
      if (r < gModel.p)
        gModel.state = LOSSY;
      break;
    case LOSSY:                         /* in lossy state */
      if (r < gModel.q)
        gModel.state = NON_LOSSY;
      break;
    default:
      /* Invalid state */
      assert(0);
      break;
  }

  updateGilbertStat(seq, gModel.state);

  return gModel.state == LOSSY ? 0 : 1;
}

#endif
