#include "defines.h"
#if FBIR

#include <pthread.h>
#include <strings.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "rtcp.h"
#include "global.h"
#include "fbir.h"

static pthread_t rtcp_thr;

#if RTCP_DEBUG_ENABLE

#include <sys/timeb.h>

static inline double timenow(void)
{
  struct timeb t;
  ftime(&t); 
  return t.time + (double) t.millitm / 1000;
}

#define RTCP_DEBUG(s) printf("%.3f: ", timenow()); printf s

#else

#define RTCP_DEBUG(s)

#endif

/*
 * Open the RTCP connection and prepare for listening
 */
static void rtcp_open_connection (uint16_t port, rtcp_connection_t *con)
{
  /* Create a socket and bind it */
  
  bzero(&con->encaddr, sizeof(con->encaddr));
  con->encaddr.sin_family = AF_INET;
  con->encaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  con->encaddr.sin_port = htons(port);

  if ((con->sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) 
  {
    perror("Server: Failed creating socket");
    exit(-1);
  }

  if ((bind(con->sockfd, (struct sockaddr *)&con->encaddr, sizeof(con->encaddr)) < 0)) 
  {
    perror("Server: Failed bind");
    exit(-1);
  }

  /* Prepare for using 'select' when listening */
  FD_ZERO(&con->readfds);
  FD_SET(con->sockfd, &con->readfds);
}


/*
 * Free all memory for RTCP packet
 */
static void rtcp_packet_free (void *pck)
{
  rtcp_packet_t *p = (rtcp_packet_t *) pck;
  
  if (p == NULL)
    return;

  if (p->rr) {
    if (p->rr->rb)
      free (p->rr->rb);
    free(p->rr);
  }
  if (p->sdes->chunk_item_cname)
    free(p->sdes->chunk_item_cname);
  if (p->sdes)
    free(p->sdes);
  if (p->fb) {
    if (p->fb->n_fci != 0) {
      free(p->fb->fci);
      p->fb->n_fci = 0;
    }
    free(p->fb);
  }
  free(p);
}


/*
 * Reset all RTCP packet variables between each packet
 */
static void rtcp_packet_reset(rtcp_packet_t *p)
{
  bzero(p->packet, p->packlen);

  if (p->fb->n_fci != 0) {
    free(p->fb->fci);
    p->fb->n_fci = 0;
  }
  p->has_fb = 0;
  p->has_sdes = 0;
  p->has_rr = 0;
  bzero(p->rr->rb, sizeof(rtcp_rr_report_block_t));
  p->packlen = 0;
}


/*
 * Allocate memory for RTCP packet
 */
static void rtcp_packet_alloc (rtcp_packet_t **pp)
{
  rtcp_packet_t *p;
  
  if ((p = (rtcp_packet_t *) calloc(1, sizeof(rtcp_packet_t))) == NULL) {
    perror("rtcp_packet_alloc() calloc() #1");
    exit(-2);
  }
  
  if ((p->rr = (rtcp_rr_t *) calloc(1, sizeof(rtcp_rr_t))) == NULL) {
    perror("rtcp_packet_alloc() calloc() #2");
    exit(-2);
  }

  if ((p->rr->rb = (rtcp_rr_report_block_t *) calloc(1, sizeof(rtcp_rr_report_block_t))) == NULL) {
    perror("rtcp_packet_alloc() calloc() #3");
    exit(-2);
  }

  if ((p->sdes = (rtcp_sdes_t *) calloc(1, sizeof(rtcp_sdes_t))) == NULL) {
    perror("rtcp_packet_alloc() calloc() #4");
    exit(-2);
  }

  if ((p->sdes->chunk_item_cname = (char *) calloc(RTCP_MAX_SDES_ITEM_LENGTH+1, sizeof(char))) == NULL) {
    perror("rtcp_packet_alloc() calloc() #5");
    exit(-2);
  }
  

  if ((p->fb = (rtcp_fb_t *) calloc(1, sizeof(rtcp_fb_t))) == NULL) {
    perror("rtcp_packet_alloc() calloc() #6");
    exit(-2);
  }

  p->fb->n_fci = 0;
  
  if ((p->packet = (byte *) calloc(RTCP_MAX_PACKET_SIZE, sizeof(byte))) == NULL) {
    perror("rtcp_packet_alloc() calloc() #7");
    exit(-2);
  }
  
  *pp = p;
}

/*
 * Listen infinitely for incoming packets (expecting RTPC packets)
 */
static int rtcp_receive_packet (rtcp_packet_t *p, rtcp_connection_t *con)
{
  int nfd;
  socklen_t socklen;

  bzero(&con->decaddr, sizeof(con->decaddr));
  socklen = sizeof(con->decaddr);

  /* Check when socket is ready for reading, infinite timeout */
  nfd = select(con->sockfd+1, &con->readfds, NULL, NULL, NULL);

  //Receive data and create the packet structure
  if (nfd > 0) {
    p->packlen = recvfrom(con->sockfd, p->packet, RTCP_MAX_PACKET_SIZE, 0,
                          (struct sockaddr *) &con->decaddr, &socklen);
    if (p->packlen < 0) {
      fprintf(stderr, "An error occured receiving RTCP data\n");
      return -1;
    }
  }
  else if (nfd == 0) {
    /* Timeout, should not happen because of infinite timeout */
    RTCP_DEBUG(("Timeout, something weird happened\n"));
    return 0;    
  }
  else {
    fprintf(stderr, "An error occured while listening for RTCP packets\n");
    return -1;
  }

  return p->packlen;
}

/*
 * Decompose RR. Support for only one Report block.
 */
static int rtcp_decompose_rr (rtcp_rr_t *rr, const byte *buf)
{
  assert(rr != NULL);
  assert(buf != NULL);

  rr->v = (buf[0] >> 6) & 0x03;
  rr->p = (buf[0] >> 5) & 0x01;
  rr->rc = (buf[0] >> 0) & 0x1F;
  assert(rr->rc == 1);                  /* Haven't implemented support for more */

  rr->pt = buf[1];
  assert(rr->pt == RTCP_RR);

  memcpy(&rr->length, &buf[2], 2);
  rr->length = ntohs(rr->length);
  assert(rr->length == 7);

  memcpy(&rr->ssrc, &buf[4], 4);
  rr->ssrc = ntohl(rr->ssrc);

  /* Report block 1 */
  memcpy(&rr->rb->ssrc_n, &buf[8], 4);
  rr->rb->ssrc_n = ntohl(rr->rb->ssrc_n);

  rr->rb->fl = buf[12];
  
  memcpy(&rr->rb->cnlp, &buf[13], 3);
  rr->rb->cnlp = ntohl(rr->rb->cnlp << 8);

  memcpy(&rr->rb->ehsn, &buf[16], 4);
  rr->rb->ehsn = ntohl(rr->rb->ehsn);

  memcpy(&rr->rb->jitter, &buf[20], 4);
  rr->rb->jitter = ntohl(rr->rb->jitter);

  memcpy(&rr->rb->lsr, &buf[24], 4);
  rr->rb->lsr = ntohl(rr->rb->lsr);

  memcpy(&rr->rb->dlsr, &buf[28], 4);
  rr->rb->dlsr = ntohl(rr->rb->dlsr);
  
  return 0;
}

/*
 * Decompose SDES. Limited support for ITEMS. Only CNAME
 */
static int rtcp_decompose_sdes (rtcp_sdes_t *sdes, byte *buf)
{
  assert(sdes != NULL);
  assert(buf != NULL);

  sdes->v = (buf[0] >> 6) & 0x03;
  sdes->p = (buf[0] >> 5) & 0x01;
  sdes->sc = (buf[0] >> 0) & 0x1F;

  sdes->pt = buf[1];
  
  memcpy(&sdes->length, &buf[2], 2);
  sdes->length = ntohs(sdes->length);

  /* Haven't implemented support for others than the following: */
  assert(sdes->v == 2);
  assert(sdes->p == 0);
  assert(sdes->sc == 1);
  assert(sdes->pt == RTCP_SDES);

  memcpy(&sdes->chunk_csrc, &buf[4], 4);
  sdes->chunk_csrc = ntohl(sdes->chunk_csrc);

  sdes->chunk_item_type = buf[8];
  sdes->chunk_item_length = buf[9];

  memcpy(sdes->chunk_item_cname, &buf[10], sdes->chunk_item_length);
  sdes->chunk_item_cname[sdes->chunk_item_length] = '\0';
  
  return 0;
}

/*
 * Returns a negative value if error occurs while decomposing FB message
 */
static int rtcp_decompose_fb (rtcp_fb_t *fb, byte *buf)
{
  int i;
  
  assert(fb != NULL);
  assert(buf != NULL);

  fb->v = (buf[0] >> 6) & 0x03;
  fb->p = (buf[0] >> 5) & 0x01;
  fb->fmt = (buf[0] >> 0) & 0x1F;
  assert(fb->v == 2);
  assert(fb->p == 0);
  assert(fb->fmt == 1);

  fb->pt = buf[1];
  assert(fb->pt == RTCP_RTPFB);
  
  memcpy(&fb->length, &buf[2], 2);
  fb->length = ntohs(fb->length);

  memcpy(&fb->ssrc_sender, &buf[4], 4);
  fb->ssrc_sender = ntohl(fb->ssrc_sender);

  memcpy(&fb->ssrc_source, &buf[8], 4);
  fb->ssrc_source = ntohl(fb->ssrc_source);

  fb->n_fci = fb->length - 2;

  if (fb->n_fci > 0) {
    if ((fb->fci = (rtcp_fci_nack_t *) malloc (sizeof(rtcp_fci_nack_t) * fb->n_fci)) == NULL) {
      perror("Error allocating memory for FCIs");
      return -1;
    }
  }
    
  for (i = 0; i < fb->n_fci; i++) {
    memcpy(&fb->fci[i].pid, &buf[12 + 4*i], 2);
    fb->fci[i].pid = ntohs(fb->fci[i].pid);

    memcpy(&fb->fci[i].blp, &buf[14 + 4*i], 2);
    fb->fci[i].blp = ntohs(fb->fci[i].blp);
  }

  return 0;
}

/*
 * Returns a negative value if there was an error while decomposing the
 * packet.
 */
static int rtcp_decompose_packet(rtcp_packet_t *p)
{
  int readlen = 0;
  uint8_t pt;
  
  assert(p->packlen < 65536 - 28);      // IP, UDP headers
  assert(p->packlen >= 4);              // Only one SDES header
  assert(p->packet != NULL);


  while (readlen < p->packlen) {
//    memcpy(&pt, &p->packet[1], 1);
    pt = (uint8_t) p->packet[readlen + 1];
    switch (pt) {
      case RTCP_RR:
        if (rtcp_decompose_rr(p->rr, &p->packet[readlen]) < 0) {
          fprintf(stderr, "Error decomposing RR\n");
          return -1;
        }
        p->has_rr = 1;
        readlen += (p->rr->length+1) * 4;
        break;
        
      case RTCP_SDES:
        if (rtcp_decompose_sdes(p->sdes, &p->packet[readlen]) < 0) {
          fprintf(stderr, "Error decomposing SDES\n");
          return -1;
        }
        p->has_sdes = 1;
        readlen += (p->sdes->length+1) * 4;
        break;

      case RTCP_RTPFB:
        if (rtcp_decompose_fb(p->fb, &p->packet[readlen]) < 0) {
          fprintf(stderr, "Error decomposing RTPFB\n");
          return -1;
        }
        p->has_fb = 1;
        readlen += (p->fb->length+1) * 4;
        break;

      default:
        fprintf(stderr, "RTCP packet type %d is not currently supported\n", pt);
        return -1;
    }
  }

//  RTCP_DEBUG(("Done decomposing, totlen = %d, packlen = %d\n", readlen, p->packlen));
  
  return 0;
}

#if 0
static void rtcp_dump_rr(rtcp_rr_t *rr)
{
  assert(rr);
  assert(rr->rb);
  assert(rr->rc == 1);
  assert(rr->length == 7);
  
  printf("RR:\n");
  printf("\tversion: %u\n", rr->v);
  printf("\tpadding: %u\n", rr->p);
  printf("\trr count: %u\n", rr->rc);
  printf("\tpacket type: %u\n", rr->pt);
  printf("\tlength: %u\n", rr->length);
  printf("\tssrc: %u\n", rr->ssrc);

  printf("\tReport block 1:\n");
  printf("\t\tssrc_1: %u\n", rr->rb->ssrc_n);
  printf("\t\tfraction: %u/256\n", rr->rb->fl);
  printf("\t\tcum lost: %u\n", rr->rb->cnlp);
  printf("\t\tmax seq: %u\n", rr->rb->ehsn);
  printf("\t\tjitter: %u\n", rr->rb->jitter);
  printf("\t\tlasr sr: %u\n", rr->rb->lsr);
  printf("\t\tdelay since sr: %u\n", rr->rb->dlsr);
}


static void rtcp_dump_sdes (rtcp_sdes_t *sdes)
{
  assert(sdes);
  assert(sdes->sc == 1);
  assert(sdes->chunk_item_type == 1);

  printf("SDES:\n");
  printf("\tversion: %u\n", sdes->v);
  printf("\tpadding: %u\n", sdes->p);
  printf("\tsdes count: %u\n", sdes->sc);
  printf("\tpacket type: %u\n", sdes->pt);
  printf("\tlength: %lu\n", (unsigned long) sdes->length);
  printf("\tchunk csrc: %u\n", sdes->chunk_csrc);
  printf("\tchunk item type: %u\n", sdes->chunk_item_type);
  printf("\tchunk item length: %u\n", sdes->chunk_item_length);
  printf("\tchunk item cname: %s", sdes->chunk_item_cname);
}

static void rtcp_dump_fb (rtcp_fb_t *fb)
{
  int i;
  
  assert(fb);

  printf("RTPFB:\n");
  printf("\tversion: %u\n", fb->v);
  printf("\tpadding: %u\n", fb->p);
  printf("\tfeedback message type: %u\n", fb->fmt);
  printf("\tpacket type: %u\n", fb->pt);
  printf("\tlength: %u\n", fb->length);
  printf("\tssrc sender: %u\n", fb->ssrc_sender);
  printf("\tssrc source: %u\n", fb->ssrc_source);
  
  for (i = 0; i < fb->n_fci; i++) {
    printf("\tFCI [%d]:\n", i);
    printf("\t\tpid: %u\n", fb->fci[i].pid);
    printf("\t\tblp: 0x%04X\n", fb->fci[i].blp);
  }
}


static void rtcp_dump_packet(rtcp_packet_t *p)
{
  assert(p);

  printf("===============================================\n");
  if (p->has_rr) {
    rtcp_dump_rr(p->rr);
    printf("\n");
  }
  if (p->has_sdes) {
    rtcp_dump_sdes(p->sdes);
    printf("\n");
  }
  if (p->has_fb) {
    rtcp_dump_fb (p->fb);
    printf("\n");
  }
}
#endif


/*
 * Interpret RTCP packet and update according to the FB information.
 *
 * Returns 0 if OK.
 */
static int rtcp_interpret (rtcp_packet_t *p)
{
  int pid;
  int i, j;

  /* Update network statistics based on RR */
  if (p->has_rr) {
    feedback->lossrate_last = (feedback->lossrate_last + 1) % LOSSRATE_MEMORY_SIZE;
    feedback->lossrate[feedback->lossrate_last] = p->rr->rb->fl * 100.0 / 256;
    feedback->lossrate_total = p->rr->rb->cnlp * 100.0 / p->rr->rb->ehsn;
  }
   
  /* If an FB message indicate lost packets, perform error tracking (or more
   * accurately, notify other parts of the encoder about the loss and let them
   * perform the tracking) */
  if (p->has_fb) {
    assert(p->fb->fmt == 1);               /* Only support NACK */

    for (i = 0; i < p->fb->n_fci; i++) {
      pid = p->fb->fci[i].pid;
      register_lost_packet(pid);

      for (j = 0; j < 16; j++) {
        if ((p->fb->fci[i].blp >> j) & 0x01)
          register_lost_packet(pid + j + 1);
      }
    }
  }

//  rtcp_dump_packet(p);
  return 0;
}

/*
 * This is the main loop for RTCP. Should be invoked through a pthread.
 * It listens on defined port and start the processing of the incoming RTCP
 * packets.
 */
static void *rtcp_main (void *arg)
{
  int len;
  rtcp_packet_t *p;
  rtcp_connection_t con;
  unsigned short port = input->feedbackPort;
  
  rtcp_packet_alloc(&p);

  pthread_cleanup_push(rtcp_packet_free, (void *) p);
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

  rtcp_open_connection(port, &con);

  /* Start listen, decompose report the incoming messages */
  while ((len = rtcp_receive_packet(p, &con)) > 0) {
    if (rtcp_decompose_packet(p) < 0) {
      fprintf(stderr, "Error decomposing RTCP packet, exit\n");
      exit(-1);
    }
    if (rtcp_interpret (p) < 0) {
      fprintf(stderr, "Error interpreting RTCP packet, exit\n");
      exit(-1);
    }
    rtcp_packet_reset(p);
  }

  pthread_cleanup_pop(1);
  pthread_exit(0);
}


/*
 * Creates a socket that (only) listens for incoming RTCP feedback messages on
 * given port.
 */
void OpenRTCPConnection(unsigned short port)
{
  pthread_attr_t attr;
    
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    
  pthread_create(&rtcp_thr, NULL, rtcp_main, &port);

  pthread_attr_destroy(&attr);
}


/*
 * Close the RTCP connection.
 */
void CloseRTCPConnection()
{
  pthread_cancel(rtcp_thr);
}

#endif
